import numpy as np
import itertools
import matplotlib.pyplot as plt
from kawin.precipitation.non_ideal.LebedevNodes import loadPoints
import copy

class StrainEnergy:
    '''
    Defines class for calculating strain energy of a precipitate

    Ellipsoidal precipitates will use the Eshelby's tensor
    Spherical and Cuboidal precipitates will use the Khachaturyan's approximation
    '''

    SPHERE = 0
    ELLIPSE = 1
    CUBE = 2
    CONSTANT = 3

    def __init__(self):
        self.c = None
        self._c4 = np.zeros((3,3,3,3))
        self.cPrec = None
        self._c4Prec = np.zeros((3,3,3,3))
        self.eigstrain = np.zeros((3,3))
        self.appstress = np.zeros((3,3))
        self.appstrain = np.zeros((3,3))
        self.rotation = None
        self.rotationPrec = None

        self.lebedevIntegration('high')

        self._strainEnergyGeneric = self._strainEnergyConstant
        self.type = self.CONSTANT
        self.r = np.zeros(3)
        self.oldr = np.zeros(3)
        self._prevEnergy = 0

        self._gElasticConstant = 0

        #Cached values for calculating equilibrium aspect ratio
        self.ifmethod = 1
        self._aspectRatios = None
        self._normEnergies = None
        self._cachedRange = 5
        self._cachedIntervals = 100

        self._ohm_inverse = self._ohm_quickInverse

    def setOhmInverseFunction(self, method = 'quick'):
        '''
        Sets method to invert the ohm term in calculating eshelby's tensor

        Parameters
        ----------
        method : str
            'numpy' - uses np.linalg.inv, which can be slower for batch, but runs through
                        multiple checks for whether values are real/complex or if inverse exists
            'quick' - quick inverse using Cramer's rule assuming that values are real and inverse exists - recommended method
        '''
        if method == 'numpy':
            self._ohm_inverse = self._ohm_npinv
        else:
            self._ohm_inverse = self._ohm_quickInverse

    def setAspectRatioResolution(self, resolution = 0.01, cachedRange = 5):
        '''
        Sets resolution to which equilibrium aspect ratios are calculated

        Equilibrium aspect ratios are found by calculated strain energy for a range of aspect ratios
        and finding the aspect ratio giving the minimum energy (strain + interfacial energy)

        If aspect ratio does not vary much in a given system, then the default parameters may lead to poor
        prediction of the aspect ratios

        Parameters
        ----------
        resolution : float (optional)
            Minimum distance between aspect ratios when calculating cache (default at 0.01)
        cachedRange : float (optional)
            Range of aspect ratio to calculate strain energy when updated cache (default at 5)
        '''
        self._cachedRange = cachedRange
        self._cachedIntervals = int(1 / resolution)

    def setInterfacialEnergyMethod(self, method):
        '''
        Sets method for calculating interfacial energy as a function of aspect ratio

        Parameters
        ----------
        method : str
            'eqradius' - interfacial energy is determined using the equivalent spherical radius
            'thermo' - interfacial energy is determined using dG/dSA (default)
        '''
        if method == 'eqradius':
            self.ifmethod = 0
        else:
            self.ifmethod = 1

    def setSpherical(self):
        '''
        Assumes spherical geometry for strain energy calculation
        Uses Khachaturyan's approximation
        '''
        self._strainEnergyGeneric = self._strainEnergySphere
        self.type = self.SPHERE
    
    def setCuboidal(self):
        '''
        Assumes cuboidal geometry for strain energy calculation
        Uses Khachaturyan's approximation
        '''
        self._strainEnergyGeneric = self._strainEnergyCube
        self.type = self.CUBE

    def setEllipsoidal(self):
        '''
        Assumes ellipsoidal geometry for strain energy calculation
        Uses Eshelby's tensor
        '''
        self._strainEnergyGeneric = self._strainEnergyEllipsoid
        self._strainEnergyGeneric = self._strainEnergyBohm
        self.type = self.ELLIPSE

    def setConstantElasticEnergy(self, energy):
        '''
        If elastic strain energy is known to be a constant, this can be use to greatly
        simplify calculations

        Parameters
        ----------
        energy - strain energy in J/m3
        '''
        self._gElasticConstant = energy
        self._strainEnergyGeneric = self._strainEnergyConstant
        self.type = self.CONSTANT

    def setElasticTensor(self, tensor):
        '''
        Sets elastic tensor of matrix using 2nd rank tensor

        Parameters
        ----------
        tensor : 6x6 array
            2nd rank elastic tensor
        '''
        self.c = self._setElasticTensor(tensor)
        self._c4 = self._convert2To4rankTensor(self.c)

    def setElasticConstants(self, c11, c12, c44):
        '''
        Sets elastic tensor of matrix by elastic constants, assuming cubic symmetry

        Parameters
        ----------
        c11 : float
            Modulus for compression
            c11 = E(1-nu) / (1+nu)(1-2nu)
        c12 : float
            Modulus for dilation (accounts for compression and Poisson's ratio)
            c12 = E nu / (1+nu)(1-2nu)
        c44 : float
            Modulus for shear
            c44 = (c11-c12)/2
        '''
        self.c = self._setElasticConstants(c11, c12, c44)
        self._c4 = self._convert2To4rankTensor(self.c)

    def setModuli(self, E = None, nu = None, G = None, lam = None, K = None, M = None):
        '''
        Sets elastic tensor of matrix by 2 moduli

        Parameters (only 2 has to be defined)
        ----------
        E : float
            Elastic modulus
        nu : float
            Poisson's ratio
        G : float
            Shear modulus
        lam : float
            Lame's first parameter
        K : float
            Bulk modulus
        M : float
            P-wave modulus
        '''
        self.c = self._setModuli(E, nu, G, lam, K, M)
        self._c4 = self._convert2To4rankTensor(self.c)

    def setElasticTensorPrecipitate(self, tensor):
        '''
        Sets elastic tensor of precipitate using 2nd rank tensor

        Parameters
        ----------
        tensor : 6x6 array
            2nd rank elastic tensor
        '''
        self.cPrec = self._setElasticTensor(tensor)
        self._c4Prec = self._convert2To4rankTensor(self.cPrec)

    def setElasticConsantsPrecipitate(self, c11, c12, c44):
        '''
        Sets elastic tensor of precipitate by elastic constants, assuming cubic symmetry

        Parameters
        ----------
        c11 : float
            Modulus for compression
            c11 = E(1-nu) / (1+nu)(1-2nu)
        c12 : float
            Modulus for dilation (accounts for compression and Poisson's ratio)
            c12 = E nu / (1+nu)(1-2nu)
        c44 : float
            Modulus for shear
            c44 = (c11-c12)/2
        '''
        self.cPrec = self._setElasticConstants(c11, c12, c44)
        self._c4Prec = self._convert2To4rankTensor(self.cPrec)
    
    def setModuliPrecipitate(self, E = None, nu = None, G = None, lam = None, K = None, M = None):
        '''
        Sets elastic tensor of precipitate by 2 moduli

        Parameters (only 2 has to be defined)
        ----------
        E : float
            Elastic modulus
        nu : float
            Poisson's ratio
        G : float
            Shear modulus
        lam : float
            Lame's first parameter
        K : float
            Bulk modulus
        M : float
            P-wave modulus
        '''
        self.cPrec = self._setModuli(E, nu, G, lam, K, M)
        self._c4Prec = self._convert2To4rankTensor(self.cPrec)

    def _setElasticTensor(self, tensor):
        '''
        Inputs 6x6 elastic matrix

        Parameters
        ----------
        tensor : matrix of floats
            Must be 6x6
        '''
        return np.array(tensor)

    def _setElasticConstants(self, c11, c12, c44):
        '''
        Creates elastic tensor from c11, c12 and c44 constants assuming isotropic system

        Parameters
        ----------
        c11 : float
        c12 : float
        c44 : float
        '''
        c = np.zeros((6, 6))
        c[0,0], c[1,1], c[2,2] = c11, c11, c11
        c[0,1], c[0,2], c[1,0], c[1,2], c[2,0], c[2,1] = c12, c12, c12, c12, c12, c12
        c[3,3], c[4,4], c[5,5] = c44, c44, c44
        return c

    def _setModuli(self, E = None, nu = None, G = None, lam = None, K = None, M = None):
        '''
        Set elastic tensor by defining at least two of the moduli
        Everything will be converted to E, nu and G
        If more than two parameters are defined, then the following priority will be used:
        E - Youngs modulus
        nu - Poisson's ratio
        G - shear modulus
        lam - Lame's first parameter
        K - bulk modulus
        M - P-wave modulus

        NOTE: There's gotta be a better way to implement the conversions
        '''
        if E:
            if nu:
                G = E / (2 * (1 + nu))
            elif G:
                nu = E / (2 * G) - 1
            elif lam:
                R = np.sqrt(E**2 + 9*lam**2 + 2*E*lam)
                nu = 2*lam / (E + lam + R)
                G = (E - 3*lam + R) / 4
            elif K:
                nu = (3*K - E) / (6*K)
                G = 3*K*E / (9*K - E)
            elif M:
                S = np.sqrt(E**2 + 9*M**2 - 10*E*M)
                nu = (E - M + S) / (4*M)
                G = (3*M + E - S) / 8
        elif nu:
            if G:
                E = 2*G*(1 + nu)
            elif lam:
                E = lam * (1 + nu) * (1 - 2*nu) / nu
                G = lam * (1 - 2*nu) / (2*nu)
            elif K:
                E = 3*K*(1 - 2*nu)
                G = 3*K*(1 - 2*nu) / (2*(1 + nu))
            elif M:
                E = M * (1 + nu) * (1 - 2*nu) / (1 - nu)
                G = M * (1 - 2*nu) / (2 * (1 - nu))
        elif G:
            if lam:
                E = G * (3*lam + 2*G) / (lam + G)
                nu = lam / (2*(lam + G))
            elif K:
                E = 9*K*G / (3*K + G)
                nu = (3*K - 2*G) / (2*(3*K + G))
            elif M:
                E = G * (3*M - 4*G) / (M - G)
                nu = (M - 2*G) / (2*M - 2*G)
        elif lam:
            if K:
                E = 9*K*(K - lam) / (3*K - lam)
                G = 3*(K - lam) / 2
                nu = lam / (3*K - lam)
            elif M:
                E = (M - lam) * (M + 2*lam) / (M + lam)
                G = (M - lam) / 2
                nu = lam / (M + lam)
        elif K:
            if M:
                E = 9*K*(M - K) / (3*K + M)
                G = 3*(M - K) / 4
                nu = (3*K - M) / (3*K + M)

        s = np.zeros((6,6))
        s[0,0], s[1,1], s[2,2] = 1/E, 1/E, 1/E
        s[3,3], s[4,4], s[5,5] = 1/G, 1/G, 1/G
        s[0,1], s[0,2], s[1,0], s[1,2], s[2,0], s[2,1] = -nu/E, -nu/E, -nu/E, -nu/E, -nu/E, -nu/E
        return np.linalg.inv(s)

    def setRotationMatrix(self, rot):
        '''
        Sets rotation matrix to be applied to the matrix

        This is for cases where the axes of the precipitate does not align with the axes of the matrix
        (e.g., the long/short axes of the precipitate is not parallel to the <100> directions of the matrix)

        Parameters
        ----------
        rot : matrix
            3x3 rotation matrix
        '''
        self.rotation = np.array(rot)

    def setRotationPrecipitate(self, rot):
        self.rotationPrec = np.array(rot)

    def setEigenstrain(self, strain):
        '''
        Sets eigenstrain of precipitate

        Parameters
        ----------
        strain : float, array or matrix
            float - assume strain is the same along all 3 axis
            array - each index corresponds to strain in a given axis
            matrix - full 2nd rank strain tensor

        NOTE: when using in conjunction with ShapeFactors, the axis are order specific
            For needle-like precipitates, x1, x2 and x3 correspond to (short axis, short axis, long axis)
            For plate-like precipitates, x1, x2 and x3 correspond to (long axis, long axis, short axis)
        '''
        strain = np.array(strain)
        #If scalar, then apply to all 3 axis
        if (type(strain) == np.ndarray and strain.ndim == 0):
            self.eigstrain = strain * np.identity(3)
        #If array of length 3, then apply strain along each index to corresponding axis
        elif strain.ndim == 1:
            self.eigstrain = np.array([[strain[0], 0, 0], [0, strain[1], 0], [0, 0, strain[2]]])
        #Else, assume it's a tensor
        else:
            self.eigstrain = strain

    def setAppliedStress(self, stress):
        '''
        Sets applied stress tensor
        Axes of stress tensor should be the same as the matrix

        Parameters
        ----------
        stress : float, array or matrix
            float - assume stress is the same along all 3 axis
            array - each index corresponds to stress in a given axis
            matrix - full 2nd rank stress tensor

        NOTE: The applied stress is in reference to the coordinate axes of the precipitates
        Thus, this is only valid if the following conditions are met:
            The matrix phase has a single orientation (either as a single crystal or highly textured)
            Precipitates will form only in a single orientation with respect to the matrix

        TODO: It will be nice to expand on this.
            For polycrystalline matrix and randomly oriented precipitates, it should be possible
            to average the strain energy contributions over all matrix/precipitate orientations
        '''
        stress = np.array(stress)
        #If scalar, then apply to all 3 axis
        if (type(stress) == np.ndarray and stress.ndim == 0):
            self.appstress = stress * np.identity(3)
        #If array of length 3, then apply stress along each index to corresponding axis
        elif stress.ndim == 1:
            self.appstress = np.array([[stress[0], 0, 0], [0, stress[1], 0], [0, 0, stress[2]]])
        #Else, assume it's a tensor
        else:
            self.appstress = stress

    def getAppliedStrain(self):
        '''
        Calculates applied strain tensor from applied stress tensor and elastic tensor
        '''
        flatStress = self._convert2rankToVec(self.appstress)
        flatStrain = np.matmul(np.linalg.inv(self.c), flatStress)
        self.appstrain = self._convertVecTo2rankTensor(flatStrain)

    def setup(self):
        '''
        Sets up elastic constants
        '''
        #If no elastic tensor is given, then elastic energy will be constant at 0
        if self.c is not None:
            #Apply rotation on matrix phase if there is one
            #Since applied stress tensor is aligned with matrix, rotate it too since calculations will be in reference to precipitate axes
            if self.rotation is not None:
                self._c4 = self._rotateRank4Tensor(self.rotation, self._c4)
                self.c = self._convert4To2rankTensor(self._c4)
                self.appstress = self._rotateRank2Tensor(self.rotation, self.appstress)

            #If elastic constants for precipitate are not given, then assume they're the same as the matrix
            #This will give the elastic energy equation from Wu et al, Journal of Phase Equilibria and Diffusion, 39, 2018
            #Using a different elastic tensor for the precipitate will give energy from Bohm et al, Mechanics of Materials, 155, 2021
            if self.cPrec is None:
                self.cPrec = copy.copy(self.c)
                self._c4Prec = copy.copy(self._c4)

            if self.rotationPrec is not None:
                self._c4Prec = self._rotateRank4Tensor(self.rotationPrec, self._c4Prec)
                self.cPrec = self._convert4To2rankTensor(self._c4Prec)

            self.getAppliedStrain()

            #Set type to something other than CONSTANT
            #Since CONSTANT is the only method not requiring the elastic tensor,
            #    we assume that the user is intending to calculate strain energy when inputting the tensor
            if self.type == self.CONSTANT:
                self.type = self.SPHERE

    #Utility functions for tensors
    def _convert2To4rankTensor(self, c):
        '''
        Converts 2nd rank elastic tensor to 4th rank

        Parameters
        ----------
        c : ndarray
            2nd rank elastic tensor

        Returns
        -------
        c4 : ndarray
            4th rank elastic tensor
        '''
        vMap = {frozenset({0}): 0, frozenset({1}): 1, frozenset({2}): 2,
                frozenset({1,2}): 3, frozenset({0,2}): 4, frozenset({0,1}): 5}

        c4 = np.zeros((3,3,3,3))
        for i, j, k, l in itertools.product(range(3), range(3), range(3), range(3)):
            c4[i,j,k,l] = c[vMap[frozenset({i,j})], vMap[frozenset({k,l})]]
        return c4

    def _convert4To2rankTensor(self, c4):
        '''
        Converts 4th rank elastic tensor to 4nd rank

        Parameters
        ----------
        c4 : ndarray
            4th rank elastic tensor

        Returns
        -------
        c : ndarray
            2nd rank elastic tensor
        '''
        vMap = [[0,0], [1,1], [2,2], [1,2], [0,2], [0,1]]
        c = np.zeros((6,6))
        for i, j in itertools.product(range(6), range(6)):
            c[i,j] = c4[vMap[i][0], vMap[i][1], vMap[j][0], vMap[j][1]]
        return c

    def _invert4rankTensor(self, c4):
        '''
        Inverts 4th rank tensor to give stiffness tensor

        This is done by converting to 2nd rank, inverting, then converting back to 4th rank
        '''
        c = self._convert4To2rankTensor(c4)
        return self._convert2To4rankTensor(np.linalg.inv(c))

    def _convertVecTo2rankTensor(self, v):
        '''
        Converts strain/stress vector to 2nd rank tensor

        Parameters
        ----------
        v : 1d array
            Strain/stress vector

        Returns
        -------
        e : ndarray
            2nd rank elastic tensor
        '''
        return np.array([[v[0], v[5], v[4]], \
                        [v[5], v[1], v[3]], \
                        [v[3], v[4], v[2]]])

    def _convert2rankToVec(self, c):
        '''
        Converts 2nd rank tensor to vector

        Parameter
        ---------
        c : ndarray
            3x3 tensor

        Returns
        -------
        v : 1darray
            Strain/stress vector
        '''
        return np.array([c[0,0], c[1,1], c[2,2], c[1,2], c[0,2], c[0,1]])

    def _rotateRank2Tensor(self, rot, tensor):
        '''
        Rotates a 2nd rank tensor
        T_ij = r_il * r_jk * T_lk

        Parameters
        ----------
        tensor : ndarray
            2nd rank tensor to rotate (3x3 array)
        '''
        return np.tensordot(rot,
                np.tensordot(rot, tensor, axes=(1,1)), axes=(1,1))

    def _rotateRank4Tensor(self, rot, tensor):
        '''
        Rotates a 4th rank tensor
        T_ijkl = r_im * r_jn * r_ok * r_lp * T_mnop

        Parameters
        ----------
        tensor : ndarray
            4th rank tensor to rotate (3x3x3x3 array)
        '''
        return np.tensordot(rot, 
                np.tensordot(rot, 
                np.tensordot(rot, 
                np.tensordot(rot, tensor, axes=(1,3)), axes=(1,3)), axes=(1,3)), axes=(1,3))

    def setIntegrationIntervals(self, phiInt, thetaInt, assumeSymmetric=True):
        '''
        Number of intervals to split domain along phi and theta for integration

        Parameters
        ----------
        phiInt : int
            Number of intervals to divide along phi
        thetaInt : int
            Number of intervals to divide along theta
        assumeSymmetric : bool (optional)
            If True (default), will only integrate Eshelby's tensor on a single quadrant and
            multiply the results by 8
        '''
        #If assume symmetric, then only a single quadrant will be integrated over
        phiRange = np.pi/2 if assumeSymmetric else 2*np.pi
        thetaRange = np.pi/2 if assumeSymmetric else np.pi
        dphi = phiRange / phiInt
        dtheta = thetaRange / thetaInt
        midPhi = np.linspace(dphi/2, phiRange - dphi/2, phiInt)
        midTheta = np.linspace(dtheta/2, thetaRange - dtheta/2, thetaInt)
        self.dA = dtheta*dphi if assumeSymmetric else 1/8 * dtheta*dphi

        #Cartesian product of phi and theta intervals
        self.midPhiGrid, self.midThetaGrid = np.meshgrid(midPhi, midTheta)
        self.midPhiGrid = self.midPhiGrid.ravel()
        self.midThetaGrid = self.midThetaGrid.ravel()
        self.midWeights = np.sin(self.midThetaGrid)

    def lebedevIntegration(self, order = 'high'):
        '''
        Creates Lebedev quadrature points and nodes for integrating Eshebly's tensor
        This is preferred over discretizing phi and theta

        Parameters
        ----------
        order : str
            'low' - uses quadrature order or 53 (974 points)
            'mid' - uses quadrature order or 83 (2354 points)
            'high' (default) - uses quadrature order or 131 (5810 points)
        '''
        if order == 'low':
            order = 53
        elif order == 'mid':
            order = 83
        else:
            order = 131

        self.midPhiGrid, self.midThetaGrid, self.midWeights = loadPoints(order)
        self.dA = np.pi/2

    def _n(self, phi, theta):
        '''
        Unit normal vector of sphere

        Parameters
        ----------
        phi - azimuthal angle
        theta - polar angle
        '''
        return np.array([np.sin(theta)*np.cos(phi), np.sin(theta)*np.sin(phi), np.cos(theta)])
    
    def _beta(self, a, b, c, phi, theta):
        '''
        Distance from center to surface of ellipsoid

        Parameters
        ----------
        a, b, c - radii of ellipsoid along x,y,z axes
        phi, theta - azimuthal, polar angle
        '''
        return np.sqrt(((a*np.cos(phi))**2 + (b*np.sin(phi))**2)*np.sin(theta)**2 + (c*np.cos(theta))**2)

    def _OhmGeneral(self, n):
        '''
        Ohm term for general system

        Ohm_ij = inverse(C_iklj * n_k * n_l)
        '''
        invOhm = np.tensordot(self._c4, np.tensordot(n, n, axes=0), axes=[[1,2], [0,1]])
        return np.linalg.inv(invOhm)
    
    def _ohm_quickInverse(self, m):
        '''
        Hard coded inverse of m which is of shape (3,3,n)

        numpy inv is more optimized for larger matrices, but can be slower for small
        matrices such as a 2x2 or 3x3. We can take advantage of 3x3 matrices having a computable
        inverse to make it faster

        NOTE: this only works since we know that m has a shape of (3,3,n) and is only composed of real numbers

        This function can probably be a bit more efficient, but quick 
        profiling on sphInt gives around a 35x speedup compared to doing 
        np.transpose(np.linalg.inv(np.transpose(m, (2,0,1))), (1,2,0)) 
        where the slowdown was in np.linalg.inv

        For matrix:
            |  a  b  c  |
            |  d  e  f  |
            |  g  h  i  |

        Inverse is defined as:
            |  ei-fh  fg-di  dh-eg  |          |  A  B  C  |
            |  ch-bi  ai-cg  bg-ah  | / det -> |  D  E  F  | / det
            |  bf-ce  cd-af  ae-bd  |          |  G  H  I  |
            Where det = aA + bB + cC
        '''
        a, b, c, d, e, f, g, h, i = m[0,0], m[0,1], m[0,2], m[1,0], m[1,1], m[1,2], m[2,0], m[2,1], m[2,2]
        A = e*i - f*h
        B = f*g - d*i
        C = d*h - e*g
        D = c*h - b*i
        E = a*i - c*g
        F = b*g - a*h
        G = b*f - c*e
        H = c*d - a*f
        I = a*e - b*d
        det = a*A + b*B + c*C
        return np.array([[A, B, C], [D, E, F], [G, H, I]]) / det
    
    def _ohm_npinv(self, m):
        '''
        Inverts ohm term using np.linalg.inv

        numpy inverse function takes in an array of shape (m,n,n) and inverts each nxn matrix
            So we have to transpose m from (3,3,n) -> (n,3,3), then invert, then transpose (n,3,3) ->(3,3,n)
        '''
        return np.transpose(np.linalg.inv(np.transpose(m, (2,0,1))), (1,2,0))

    def sphInt(self):
        '''
        Faster version of calculating the Dijkl, which avoids
        using a for loop to iterate across phi and theta
        '''
        #Normal vector (1 x 3 x n) where n is number of integration points
        n = self._n(self.midPhiGrid, self.midThetaGrid)
        nexp = np.expand_dims(n, axis=0)

        #n_i*n_j for each grid point (n x 3 x 1) x (n x 1 x 3) = (n x 3 x 3) -> (3 x 3 x n)
        nProd = np.matmul(np.transpose(nexp, (2,1,0)), np.transpose(nexp, (2,0,1))).transpose((1,2,0))

        #End term inside integral = sin(theta) / beta**3
        #endTerm = np.sin(self.midThetaGrid) / self._beta(self.r[0], self.r[1], self.r[2], self.midPhiGrid, self.midThetaGrid)**3
        endTerm = 1 / self._beta(self.r[0], self.r[1], self.r[2], self.midPhiGrid, self.midThetaGrid)**3

        #Ohm term (Ohm_ij = inverse(C_iklj * n_k * n_l))
        #For all grid points (Ohm_ijn = inverse(C_iklj) * nProd_kln)
        invOhm = np.tensordot(self._c4, nProd, axes=[[1,2], [0,1]])

        ohm = self._ohm_inverse(invOhm)

        #Tensor product (D_ijkl = intergral(ohm_ij * n_k * n_l * endTerm))
        #For summing over grid points (D_ijkl = ohm_ij * nProd_kln * endTerm_n)
        d = np.tensordot(ohm, np.multiply(nProd, endTerm * self.midWeights), axes=[[2], [2]])

        #Multiply by differential area and across the 8 quadrants
        return 8*d*self.dA

    def Dijkl(self):
        '''
        Dijkl term for Eshelby's theory
        '''
        #return -np.prod(self.r)/(4*np.pi) * self.sphericalIntegral(self.Dfunc)
        return -np.prod(self.r)/(4*np.pi) * self.sphInt()

    def Sijmn(self, D):
        '''
        S_ijmn = -0.5 * C_lkmn * (D_iklj + D_jkli)
        '''
        S = -0.5 * np.tensordot(self._c4, D + np.transpose(D, (3,1,2,0)), axes=[[0,1],[2,1]])
        #The tensor product gives S_mnij so we'll need to transpose it
        return np.transpose(S, (2,3,0,1))

    def _multiply(self, a, b):
        '''
        Multiplies 2 tensors
        4th x 2nd -> c_ij = a_ijkl * b_kl
        4th x 4th -> c_ijkl = a_ijmn * b_mnkl
        '''
        return np.tensordot(a, b, axes=[[2,3], [0,1]])

    def _strainEnergy(self, stress, strain, V):
        '''
        u = -0.5 * V * sigma_ij * strain_ij
        '''
        return -0.5 * V * np.sum(stress * strain)

    def _strainEnergyEllipsoidWithStress(self):
        '''
        Strain energy of ellipsoidal particle with applied stress
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        S = self.Sijmn(self.Dijkl())
        stress = self._multiply(self._c4, self._multiply(S, self.eigstrain) - self.eigstrain)
        stress0 = self._multiply(self._c4, self._multiply(S, self.appstrain) - self.appstrain)
        return self._strainEnergy(stress - stress0, self.eigstrain - self.appstrain, V)

    def _strainEnergyEllipsoid(self):
        '''
        Strain energy of ellipsoidal particle
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        S = self.Sijmn(self.Dijkl())
        stress = self._multiply(self._c4, self._multiply(S, self.eigstrain) - self.eigstrain)
        return self._strainEnergy(stress, self.eigstrain, V)

    def _strainEnergyEllipsoid2(self):
        '''
        Alternative method of strain energy on ellipsoidal particle using 2nd rank tensors
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        S = self._convert4To2rankTensor(self.Sijmn(self.Dijkl()))
        eigFlat = self._convert2rankToVec(self.eigstrain)
        multTerm = np.matmul(self.c, S - np.eye(6))
        return -0.5 * V * np.matmul(eigFlat, np.matmul(multTerm, eigFlat))

    def _strainEnergyBohm(self):
        '''
        Strain energy of particle for when matrix and precipitate phases have different elastic tensors
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        S = self.Sijmn(self.Dijkl())
        #invTerm = np.linalg.tensorinv(self._multiply(self._c4Prec - self._c4, S) + self._c4)
        invTerm = self._invert4rankTensor(self._multiply(self._c4Prec - self._c4, S) + self._c4)
        multTerm = self._multiply(invTerm, self._c4Prec)
        stressC = self._multiply(self._c4, self._multiply(self._multiply(S, multTerm), self.eigstrain))
        stress0 = self._multiply(self._c4, self._multiply(multTerm, self.eigstrain))
        return self._strainEnergy(stressC-stress0, self.eigstrain, V)

    def _strainEnergyBohm2(self):
        '''
        Strain energy of particle for when matrix and precipitate phases have different elastic tensors using 2nd rank tensors
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        S = self._convert4To2rankTensor(self.Sijmn(self.Dijkl()))
        eigFlat = self._convert2rankToVec(self.eigstrain)
        invTerm = np.linalg.inv(np.matmul(self.cPrec - self.c, S) + self.c)
        multTerm = np.matmul(invTerm, self.cPrec)
        stressC = np.matmul(self.c, np.matmul(np.matmul(S, multTerm), eigFlat))
        stress0 = np.matmul(self.c, np.matmul(multTerm, eigFlat))
        return -0.5 * V * np.matmul(eigFlat, stressC - stress0)

    def _Khachaturyan(self, I1, I2):
        '''
        Khachaturyan's approximation for strain energy of spherical and cuboidal precipitates
        '''
        V = 4*np.pi/3 * np.prod(self.r)
        A1 = 2 * (self.c[0,0] - self.c[0,1]) / self.c[0,0]
        A1 -= 12 * (self.c[0,0] + 2 * self.c[0,1]) * (self.c[0,0] - self.c[0,1] - 2 * self.c[3,3]) / (self.c[0,0] * (self.c[0,0] + self.c[0,1] + 2*self.c[3,3])) * I1
        A2 = -54 * (self.c[0,0] + 2 * self.c[0,1]) * (self.c[0,0] - self.c[0,1] - 2 * self.c[3,3])**2 / (self.c[0,0] * (self.c[0,0] + self.c[0,1] + 2 * self.c[3,3]) * (self.c[0,0] + 2 * self.c[0,1] + 4 * self.c[3,3])) * I2
        return 0.5 * (self.c[0,0] + 2 * self.c[0,1]) * (A1 + A2) * self.eigstrain[0,0]**2 * V

    def _strainEnergyCube(self):
        '''
        Strain energy of perfect cube (cubic factor = sqrt(2))
        '''
        return self._Khachaturyan(0.006931, 0.000959)
    
    def _strainEnergySphere(self):
        '''
        Strain energy of perfect sphere (cubic factor = 1)
        '''
        return self._Khachaturyan(1/15, 1/105)

    def _strainEnergyCuboidal(self, eta = 1):
        '''
        For cuboidal preciptitates, strain energy can be approximated
        as a linear interpolation between a perfect cube and sphere
        '''
        sC = self.strainEnergyCube()
        sS = self.strainEnergySphere()
        return (sC - sS) * (eta - 1) / (np.sqrt(2) - 1) + sS

    def _strainEnergyConstant(self):
        return 4 * np.pi / 3 * np.prod(self.r) * self._gElasticConstant

    def _strainEnergySingle(self, rsingle):
        '''
        Generic strain energy function
        '''
        self.r = rsingle
        self._prevEnergy = self._prevEnergy if np.array_equal(self.r, self.oldr) else self._strainEnergyGeneric()
        self.oldr = self.r
        return self._prevEnergy

    def strainEnergy(self, r):
        r = np.array(r)
        #If single set of radii
        if r.ndim == 1:
            return self._strainEnergySingle(r)
        else:
            eng = np.zeros(len(r))
            for i in range(len(r)):
                eng[i] = self._strainEnergySingle(r[i])
            return eng

    #Additional methods for verifying that the general methods reduce to these
    #specific cases for cubic crystal symmetry
    def _OhmCubic(self, n):
        '''
        Ohm term for cubic crystal symmetry
        '''
        ohm = np.zeros((3,3))
        jk = [(1,2), (0,2), (0,1)]
        for i in range(3):
            for j in range(3):
                if i == j:
                    jn, k = jk[i]
                    ohm[i,i] = (self.c[3,3] + (self.c[0,0]-self.c[3,3])*(n[jn]**2 + n[k]**2) + self._xi*(self.c[0,0]+self.c[0,1])*(n[jn]*n[k])**2) / (self.c[3,3]*self._D(n))
                else:
                    k = 3 - (i + j)
                    ohm[i,j] = -(self.c[0,1] + self.c[3,3])*(1 + self._xi*n[k]**2)*n[i]*n[j] / (self.c[3,3]*self._D(n))
        
        return ohm

    def _D(self, n):
        '''
        Needed for the Ohm term with cubic crystal symmetry
        '''
        d = self.c[0,0]
        d += self._xi*(self.c[0,0] + self.c[0,1])*((n[0]*n[1])**2 + (n[0]*n[2])**2 + (n[1]*n[2])**2)
        d += self._xi**2 * (self.c[0,0] + 2*self.c[0,1] + self.c[3,3])*(n[0]*n[1]*n[2])**2
        return d

    @property
    def _xi(self):
        '''
        Needed for the Ohm term with cubic crystal symmetry
        '''
        return (self.c[0,0] - self.c[0,1] - 2*self.c[3,3]) / self.c[3,3]

    #Equilibrium aspect ratios
    #Determined by minimum of strain energy + interfacial energy
    def eqAR_byGR(self, Rsph, gamma, shpFactor, a=1.001, b=100):
        '''
        Equilibrium aspect ratio using golden ratio search

        Parameters
        ----------
        Rsph : float or array
            Equivalent spherical radius
        gamma : float
            Interfacial energy
        shpFactor : ShapeFactor object
        a, b : float
            Min and max bounds
            Default is 1.001 and 100
        '''
        normR = shpFactor._normalRadiiEquation
        interfacial = shpFactor._thermoEquation if self.ifmethod == 1 else shpFactor._eqRadiusEquation
        if hasattr(Rsph, '__len__'):
            eqAR = np.ones(len(Rsph))
            for i in range(len(Rsph)):
                eqAR[i] = self._GRsearch(Rsph[i], gamma, interfacial, normR, a, b)
        else:
            eqAR = self._GRsearch(Rsph, gamma, interfacial, normR, a, b)

        return eqAR

    def _GRsearch(self, Rsph, gamma, interfacial, normR, a, b):
        '''
        Golden ratio search for a single radius
        '''
        f = lambda ar: self.strainEnergy(normR(ar))*(4/3)*np.pi*Rsph**3 + gamma*interfacial(ar)*4*np.pi*Rsph**2
    
        tol = 1e-3
        invphi = (np.sqrt(5) - 1) / 2
        invphi2 = (3 - np.sqrt(5)) / 2
        h = b - a
        c, d = a+invphi2*h, a+invphi*h
        fc, fd = f(c), f(d)

        while abs(h) > tol:
            if fc < fd:
                b, d, fd = d, c, fc
                h *= invphi
                c = a + invphi2 * h
                fc = f(c)  
            else:
                a, c, fc = c, d, fd
                h *= invphi
                d = a + invphi * h
                fd = f(d)

        return (c+d)/2

    def updateCache(self, normR):
        '''
        Update cached calculations
        '''
        if self._aspectRatios is None:
            self._aspectRatios = np.linspace(1 + 1/self._cachedIntervals, 1+self._cachedRange, int(self._cachedRange*self._cachedIntervals))
            self._normEnergies = self.strainEnergy(normR(self._aspectRatios))
        else:
            addedAspectRatios = np.linspace(self._aspectRatios[-1] + 1/self._cachedIntervals, self._aspectRatios[-1] + self._cachedRange, int(self._cachedRange*self._cachedIntervals))
            addedNormEnergies = self.strainEnergy(normR(addedAspectRatios))
            self._aspectRatios = np.concatenate((self._aspectRatios, addedAspectRatios))
            self._normEnergies = np.concatenate((self._normEnergies, addedNormEnergies))

    def clearCache(self):
        '''
        Clear cached calculations
        '''
        self._aspectRatios = None
        self._normEnergies = None

    def eqAR_bySearch(self, Rsph, gamma, shpFactor):
        '''
        Equilibrium aspect ratio by cached search

        Parameters
        ----------
        Rsph : float or array
            Equivalent spherical radius
        gamma : float
            Interfacial energy
        shpFactor : ShapeFactor object
        '''
        normR = shpFactor._normalRadiiEquation
        interfacial = shpFactor._thermoEquation if self.ifmethod == 1 else shpFactor._eqRadiusEquation
        if hasattr(Rsph, '__len__'):
            eqAR = np.ones(len(Rsph))
            for i in range(len(Rsph)):
                eqAR[i] = self._cachedSearch(Rsph[i], gamma, interfacial, normR)
        else:
            eqAR = self._cachedSearch(Rsph, gamma, interfacial, normR)
        return eqAR

    def _cachedSearch(self, Rsph, gamma, interfacial, normR):
        '''
        Cached search for a single radius
        '''
        if self._aspectRatios is None:
            self.updateCache(normR)
        
        Vsph = 4/3*np.pi*Rsph**3
        Asph = 4*np.pi*Rsph**2
        eInter = Asph*interfacial(self._aspectRatios)*gamma
        eqAR = self._aspectRatios[np.argmin(self._normEnergies*Vsph+eInter)]

        #If eqAR is on the upper end (3/4) of the cached aspect ratios, then
        #added to the cached arrays until it's not
        while eqAR > self._aspectRatios[int(-self._cachedIntervals/4)] and self._aspectRatios[-1] < 100:
            self.updateCache(normR)
            eInter = Asph*interfacial(self._aspectRatios)*gamma
            eqAR = self._aspectRatios[np.argmin(self._normEnergies*Vsph+eInter)]

        return eqAR




import numpy as np
import matplotlib.pyplot as plt
from kawin.precipitation import PopulationBalanceModel
from kawin.solver import SolverType
from kawin.GenericModel import GenericModel
from kawin.precipitation.Plot import getTimeAxis

class GrainGrowthModel(GenericModel):
    '''
    Model for grain growth that can be coupled with the KWN model to account for Zener pinning

    Following implentation described in
    K. W, J. Jeppsson and P. Mason, J. Phase Equilib. Diffus. 43 (2022) 866-875

    Parameters
    ----------
    cMin : float (optional)
        Minimum grain size (default is 1e-10)
    cMax : float (optional)
        Maximum grain size (default is 1e-8)
    bins : int (optional)
        Initial bins (default is 150)
    minBins : int (optional)
        Minimum number of bins (default is 100)
    maxBins : int (optional)
        Maximum number of bins (default is 200)
    '''
    def __init__(self, cMin = 1e-10, cMax = 1e-8, bins = 150, minBins = 100, maxBins = 200, solverType = SolverType.RK4):
        super().__init__()
        self.pbm = PopulationBalanceModel(cMin, cMax, bins, minBins, maxBins)
        self._oldPSD, self._oldPSDbounds = np.array(self.pbm.PSD), np.array(self.pbm.PSDbounds)

        #Model parameters - these are values taken from the paper as general default values
        self.gbe = 0.5      #Grain boundary energy (J/m2)
        self.M = 1e-14      #Grain boundary mobility (m4/J-s)
        self.alpha = 1      #Correction factor (for when fitting data to the model)
        self.m, self.K = {'all': 1}, {'all': 4/3}   #Factors related to spatial distribution of precipitates

        self.solverType = solverType

        self.maxDissolution = 1e-6

        self.reset()

    def setGrainBoundaryEnergy(self, gbe):
        '''
        Parameters
        ----------
        gbe : float
            Grain boundary energy
        '''
        self.gbe = gbe

    def setGrainBoundaryMobility(self, M):
        '''
        Parameters
        ----------
        M : float
            Grain boundary mobility
        '''
        self.M = M

    def setAlpha(self, alpha):
        '''
        Correction factor

        Parameters
        ----------
        alpha : float
        '''
        self.alpha = alpha

    def setZenerParameters(self, m, K, phase='all'):
        '''
        Parameters for defining zener radius

        Zener radius is defined as
        Rz = K * r / f^m

        Parameters
        ----------
        m : float
            Exponential factor for volume fraction
        K : float
            Scaling factor
        phase : str (optional)
            Precipitate phase to apply parameters to
            Default is 'all'
        '''
        self.m[phase] = m
        self.K[phase] = K

    def LoadDistribution(self, data):
        '''
        Creates a particle size distribution from a set of data
        
        Parameters
        ----------
        data : array of floats
            Array of data to be inserted into PSD
        '''
        self.pbm.reset()
        self.pbm.PSD, self.pbm.PSDbounds = np.histogram(data, self.pbm.PSDbounds)
        self.pbm.PSD = self.pbm.PSD.astype('float')
        self.Normalize()
        self.avgR[0] = self.Rm(self.pbm.PSD)
        self._oldPSD, self._oldPSDbounds = np.array(self.pbm.PSD), np.array(self.pbm.PSDbounds)
        self.dissolutionIndex = self.pbm.getDissolutionIndex(self.maxDissolution, 0)

    def LoadDistributionFunction(self, function):
        '''
        Creates a particle size distribution from a function

        Parameters
        ----------
        function : function
            Takes in R and returns density
        '''
        self.pbm.reset()
        self.pbm.PSD = function(self.pbm.PSDsize)
        self.Normalize()
        self.avgR[0] = self.Rm(self.pbm.PSD)
        self._oldPSD, self._oldPSDbounds = np.array(self.pbm.PSD), np.array(self.pbm.PSDbounds)
        self.dissolutionIndex = self.pbm.getDissolutionIndex(self.maxDissolution, 0)

    def reset(self):
        '''
        Resets model with initially loaded grain size distribution
        '''
        self.time = np.zeros(1)
        self.avgR = np.zeros(1)
        self._z = 0
        self._growthRate = np.zeros(len(self.pbm.PSDbounds))
        self.pbm.reset()
        self.pbm.PSD, self.pbm.PSDbounds = np.array(self._oldPSD), np.array(self._oldPSDbounds)
        self.dissolutionIndex = 0

    def Rcr(self, x):
        '''
        Critical radius, grains larger than Rcr will growth while smaller grains will shrink

        Critical radius is defined so that the volume will be constant when applying the growth rate

        Parameters
        ----------
        x : np.array
            Grain size distribution corresponding to GrainGrowthModel.pbm.PSDbounds
        '''
        return self.pbm.SecondMomentFromN(x) / self.pbm.FirstMomentFromN(x)
    
    def Rm(self, x):
        '''
        Mean radius

        Parameters
        ----------
        x : np.array
            Grain size distribution corresponding to GrainGrowthModel.pbm.PSDbounds
        '''
        return np.cbrt(self.pbm.ThirdMomentFromN(x) / self.pbm.ZeroMomentFromN(x))
    
    def grainGrowth(self, x):
        '''
        Grain growth model
        dRi/dt = alpha * M * gbe * (1/Rcr - 1/Ri)

        Parameters
        ----------
        x : np.array
            Grain size distribution corresponding to GrainGrowthModel.pbm.PSDbounds
        '''
        return self.alpha * self.M * self.gbe * (1 / self.Rcr(x) - 1 / self.pbm.PSDbounds)
    
    def Normalize(self):
        '''
        Normalize PSD to have a third moment of 1

        Ideally, this isn't needed since the grain growth model accounts for constant volume
            But numerical errors will lead to small changes in volume over time
        '''
        self.pbm.PSD *= 1 / self.pbm.ThirdMoment()

    def constrainedGrowth(self, growthRate, z = 0):
        '''
        Constrain growth rate due to zener pinning

        The growth rate given the zener radius is defined by:
            dR/dt = alpha * M * gbe * ((1/Rcr - 1/Ri) +/- 1/Rz)
            Where 1/Rz is added if (1/Rcr - 1/Ri) + 1/Rz < 0 (inhibits grain dissolution)
            And   1/Rz is subtracted in (1/Rcr - 1/Ri) - 1/Rz) > 0 (inhibits grain growth)
            And   dR/dt is 0 for Ri between these two limits
        
        Note: Rather than Rz (zener radius), we use z here which represents the drag force
            But these are related by z = 1/Rz

        Parameters
        ----------
        growthRate : array
            Growth rate for grain sizes
        z : float (optional)
            Zener radius, default is 0, which will not change the growth rate
        '''
        upper = growthRate + self.alpha * self.M * self.gbe * z
        lower = growthRate - self.alpha * self.M * self.gbe * z
        growIndices = lower > 0
        dissolveIndices = upper < 0
        cG = np.zeros(len(growthRate))
        cG[growIndices] = lower[growIndices]
        cG[dissolveIndices] = upper[dissolveIndices]
        return cG
    
    def getCurrentX(self):
        '''
        Returns current time and grain size distribution
        '''
        return self.time[-1], [self.pbm.PSD]
    
    def getdXdt(self, t, x):
        '''
        Returns dn_i/dt for the grain size distribution

        Steps:
            1. Get grain growth rate and corrected it with zener drag force
            2. Get dn_i/dt from the PBM given the Eulerian implementation
        '''
        self._growthRate = self.grainGrowth(x[0])
        self._growthRate = self.constrainedGrowth(self._growthRate, self._z)
        return [self.pbm.getdXdtEuler(self._growthRate, 0, 0, x[0])]

    def correctdXdt(self, dt, x, dXdt):
        '''
        Corrects dn_i/dt with the new time step
        '''
        dXdt[0] = self.pbm.correctdXdtEuler(dt, self._growthRate, 0, 0, x[0])

    def getDt(self, dXdt):
        '''
        Calculated a suitable dt with the growth rate and new time step
        We'll limit the max time step to the remaining time for solving
        '''
        return self.pbm.getDTEuler(self.finalTime - self.time[-1], self._growthRate, self.dissolutionIndex)

    def postProcess(self, time, x):
        '''
        Sets grain size distribution to x and record time and average grain size

        Steps:
            1. Set grain size distribution
            2. Adjust PSD size classes
            3. Remove grains below the dissolution threshold
            4. Normalize grain size distribution to 1 (should be a tiny correction factor due to step 3)
            5. Record time and average grain size
        '''
        self.pbm.UpdatePBMEuler(time, x[0])
        self.pbm.adjustSizeClassesEuler(True)
        self.dissolutionIndex = self.pbm.getDissolutionIndex(self.maxDissolution, 0)
        #self.pbm.PSD[:self.dissolutionIndex] = 0
        self.Normalize()
        self.time = np.append(self.time, time)
        self.avgR = np.append(self.avgR, self.Rm(self.pbm.PSD))
        self.updateCoupledModels()
        return [self.pbm.PSD], False
    
    def printHeader(self):
        '''
        Header string before solving
        '''
        print('Iteration\tTime(s)\t\tSim Time(s)\tGrain Size (um)')
    
    def printStatus(self, iteration, modelTime, simTimeElapsed):
        '''
        Status string that prints every n iteration
        '''
        print('{}\t\t{:.1e}\t\t{:.1f}\t\t{:.3e}'.format(iteration, modelTime, simTimeElapsed, self.avgR[-1]*1e6))

    def computeZenerRadius(self, model):
        '''
        Gets zener radius/drag force from PrecipitateModel

        Drag force is defined as z_j = f_j^m_j / (K_j * avgR_j)
            Where f_j is volume fraction for phase j
            And   avgR_j is average radius for phase j
        The total drag force is the sum of z_j over all the phases

        Parameters
        ----------
        model : PrecpitateModel
        '''
        z = np.zeros(len(model.phases))
        for p in range(len(model.phases)):
            phaseName = model.phases[p] if model.phases[p] in self.m else 'all'
            if model.avgR[model.n, p] > 0:
                z[p] += np.power(model.betaFrac[model.n, p], self.m[phaseName]) / (self.K[phaseName] * model.avgR[model.n, p])
        self._z = np.sum(z)

    def computeZenerRadiusByN(self, model, x):
        '''
        Gets zener radius/drag force from PrecipitateModel and PSD defined by x

        Drag force is defined as z_j = f_j^m_j / (K_j * avgR_j)
            Where f_j is volume fraction for phase j
            And   avgR_j is average radius for phase j
        The total drag force is the sum of z_j over all the phases

        Parameters
        ----------
        model : PrecpitateModel
        x : list[np.array]
            List of particle size distributions in model
        '''
        z = np.zeros(len(model.phases))
        for p in range(len(model.phases)):
            volRatio = model.VmAlpha / model.VmBeta[p]
            phaseName = model.phases[p] if model.phases[p] in self.m else 'all'
            Ntot = model.PBM[p].ZeroMomentFromN(x[p])
            RadSum = model.PBM[p].MomentFromN(x[p], 1)
            fBeta = np.amin([volRatio * model.GB[p].volumeFactor * model.PBM[p].ThirdMomentFromN(x[p]), 1])
            avgR = 0 if Ntot == 0 else RadSum / Ntot

            if avgR > 0:
                z[p] += np.power(fBeta, self.m[phaseName]) / (self.K[phaseName] * avgR)
            
        self._z = np.sum(z)

    def updateCoupledModel(self, model):
        '''
        Computes zener radius/drag force from the PrecipitateModel,
        Then solves the grain growth model with the time step of the PrecipitateModel

        Parameters
        ----------
        model : PrecpitateModel
        '''
        self.computeZenerRadius(model)
        self.solve(model.time[model.n] - model.time[model.n-1], solverType=self.solverType)

    def plotDistribution(self, ax, *args, **kwargs):
        '''
        Plots particle size distribution

        Parameters
        ----------
        ax : matplotlib axes
        '''
        self.pbm.PlotCurve(ax, *args, **kwargs)
        ax.set_xlabel('Grain Radius (m)')

    def plotDistributionDensity(self, ax, *args, **kwargs):
        '''
        Plots particle size distribution density

        Parameters
        ----------
        ax : matplotlib axes
        '''
        self.pbm.PlotDistributionDensity(ax, *args, **kwargs)
        ax.set_xlabel('Grain Radius (m)')

    def plotRadiusvsTime(self, ax, bounds = None, timeUnits = 's', *args, **kwargs):
        '''
        Plots average grain radius over time

        Parameters
        ----------
        ax : matplotlib axes
        '''
        timeScale, timeLabel, bounds = getTimeAxis(self, timeUnits, bounds)
        ax.plot(self.time*timeScale, self.avgR, *args, **kwargs)
        ax.set_xlabel(timeLabel)
        ax.set_ylabel('Grain Radius (m)')
        ax.set_ylim([0, 1.1*np.amax(self.avgR)])
        ax.set_xlim(bounds)


from kawin.tests.datasets import ALZR_TDB, NICRAL_TDB, ALMGSI_DB
from kawin.precipitation import PrecipitateModel, VolumeParameter
from kawin.thermo import BinaryThermodynamics, MulticomponentThermodynamics
import numpy as np
from numpy.testing import assert_allclose

AlZrTherm = BinaryThermodynamics(ALZR_TDB, ['AL', 'ZR'], ['FCC_A1', 'AL3ZR'], drivingForceMethod='tangent')
NiAlCrTherm = MulticomponentThermodynamics(NICRAL_TDB, ['NI', 'AL', 'CR'], ['FCC_A1', 'FCC_L12'], drivingForceMethod='tangent')
AlMgSitherm = MulticomponentThermodynamics(ALMGSI_DB, ['AL', 'MG', 'SI'], ['FCC_A1', 'MGSI_B_P', 'MG5SI6_B_DP', 'B_PRIME_L', 'U1_PHASE', 'U2_PHASE'], drivingForceMethod='tangent')

AlZrTherm.setDFSamplingDensity(2000)
AlZrTherm.setEQSamplingDensity(500)
NiAlCrTherm.setDFSamplingDensity(2000)
NiAlCrTherm.setEQSamplingDensity(500)
AlMgSitherm.setDFSamplingDensity(2000)
AlMgSitherm.setEQSamplingDensity(500)

def test_binary_precipitation_dxdt():
    '''
    Check flux values of arbitrary binary precipitation problem

    We spot check a few points on dxdt rather than checking the entire array

    This uses the parameters from 01_Binary_Precipitation example
    '''
    #Create model
    model = PrecipitateModel()
    bins = 75
    minBins = 50
    maxBins = 100
    model.setPBMParameters(cMin=1e-10, cMax=1e-8, bins=bins, minBins=minBins, maxBins=maxBins)

    xInit = 4e-3        #Initial composition (mole fraction)
    model.setInitialComposition(xInit)

    T = 450 + 273.15    #Temperature (K)
    model.setTemperature(T)

    gamma = 0.1         #Interfacial energy (J/m2)
    model.setInterfacialEnergy(gamma)

    D0 = 0.0768         #Diffusivity pre-factor (m2/s)
    Q = 242000          #Activation energy (J/mol)
    Diff = lambda x, T: D0 * np.exp(-Q / (8.314 * T))
    model.setDiffusivity(Diff)

    a = 0.405e-9        #Lattice parameter
    Va = a**3           #Atomic volume of FCC-Al
    Vb = a**3           #Assume Al3Zr has same unit volume as FCC-Al
    atomsPerCell = 4    #Atoms in an FCC unit cell
    model.setVolumeAlpha(Va, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)
    model.setVolumeBeta(Vb, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)

    #Average grain size (um) and dislocation density (1e15)
    model.setNucleationDensity(grainSize = 1, dislocationDensity = 1e15)
    model.setNucleationSite('dislocations')

    #Set thermodynamic functions
    model.setThermodynamics(AlZrTherm, addDiffusivity=False)

    #This roughly follows the steps in model.solve so we can get dxdt
    model.setup()

    #Replace x (which is just all 0 right now) with an arbitrary lognormal distribution
    r = model.PBM[0].PSDsize
    sigma = 0.25
    r0 = 0.1e-8
    n = 1/(r*sigma*np.sqrt(2*np.pi)) * np.exp(-np.log(r/r0)**2/(2*sigma**2))
    model.PBM[0].PSD = n

    t, x = model.getCurrentX()
    #Call calculateDependentTerms so it can recognize that we changed PSD, otherwise, it'll use the initial values
    model._calculateDependentTerms(t, x)
    dxdt = model.getdXdt(t, x)

    #Set arbitrary final time, this is done during the solve function, but we do it here since we're not using the solve function
    #  the initial guess for the time steo will be 0.01*(1.001) regardless of finalTime
    model.finalTime = 1 
    dt = model.getDt(dxdt)

    indices = [10, 20, 30]
    vals = [6773393.32259, 1919.5404124, 0.4106318]
    assert_allclose(vals, [dxdt[0][i] for i in indices], rtol=1e-3)
    assert_allclose(dt, 0.01001, rtol=1e-3)

def test_multi_precipitation_dxdt():
    '''
    Check flux values of arbitrary binary precipitation problem

    We spot check a few points on dxdt rather than checking the entire array

    This uses the parameters from 02_Multicomponent_Precipitation example
    '''
    model = PrecipitateModel(elements=['Al', 'Cr'])
    bins = 75
    minBins = 50
    maxBins = 100
    model.setPBMParameters(cMin=1e-10, cMax=1e-8, bins=bins, minBins=minBins, maxBins=maxBins)

    model.setInitialComposition([0.098, 0.083])
    model.setInterfacialEnergy(0.023)

    T = 1073
    model.setTemperature(T)

    a = 0.352e-9        #Lattice parameter
    Va = a**3           #Atomic volume of FCC-Ni
    Vb = Va             #Assume Ni3Al has same unit volume as FCC-Ni
    atomsPerCell = 4    #Atoms in an FCC unit cell
    model.setVolumeAlpha(Va, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)
    model.setVolumeBeta(Vb, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)

    #Set nucleation sites to dislocations and use defualt value of 5e12 m/m3
    #model.setNucleationSite('dislocations')
    #model.setNucleationDensity(dislocationDensity=5e12)
    model.setNucleationSite('bulk')
    model.setNucleationDensity(bulkN0=1e30)

    model.setThermodynamics(NiAlCrTherm)

    #This roughly follows the steps in model.solve so we can get dxdt
    model.setup()

    #Replace x (which is just all 0 right now) with an arbitrary lognormal distribution
    r = model.PBM[0].PSDsize
    sigma = 0.25
    r0 = 0.1e-8
    n = 1/(r*sigma*np.sqrt(2*np.pi)) * np.exp(-np.log(r/r0)**2/(2*sigma**2))
    model.PBM[0].PSD = n

    t, x = model.getCurrentX()
    #Call calculateDependentTerms so it can recognize that we changed PSD, otherwise, it'll use the initial values
    model._calculateDependentTerms(t, x)
    dxdt = model.getdXdt(t, x)

    #Set arbitrary final time, this is done during the solve function, but we do it here since we're not using the solve function
    #  the initial guess for the time steo will be 0.01*(1.001) regardless of finalTime
    model.finalTime = 1 
    dt = model.getDt(dxdt)

    indices = [10, 20, 30]
    vals = [2.837811e+08, 8.424854e+05, 2.312587e+02]
    assert_allclose(vals, [dxdt[0][i] for i in indices], rtol=1e-3)
    assert_allclose(dt, 0.01001, rtol=1e-3)

def test_multiphase_precipitation_x_shape():
    '''
    Check the flatten and unflatten behavior for Precipitate model

    For this setup:
        getCurrentX will return a array of length p with each element being an array of length bins
        flattenX will return a 1D array of length p*bins
        unflattenX should take the output of flattenX and getCurrentX to bring the (p*bins,) to [(bins,), (bins,), ...]

    This uses the parameters from 07_Homogenization_Model example
    '''
    phases = ['FCC_A1', 'MGSI_B_P', 'MG5SI6_B_DP', 'B_PRIME_L', 'U1_PHASE', 'U2_PHASE']
    model = PrecipitateModel(phases=phases[1:], elements=['MG', 'SI'])
    bins = 75
    minBins = 50
    maxBins = 100
    model.setPBMParameters(cMin=1e-10, cMax=1e-8, bins=bins, minBins=minBins, maxBins=maxBins)

    model.setInitialComposition([0.0072, 0.0057])
    model.setVolumeAlpha(1e-5, VolumeParameter.MOLAR_VOLUME, 4)

    lowTemp = 175+273.15
    highTemp = 250+273.15
    model.setTemperature(([0, 16, 17], [lowTemp, lowTemp, highTemp]))

    gamma = {
        'MGSI_B_P': 0.18,
        'MG5SI6_B_DP': 0.084,
        'B_PRIME_L': 0.18,
        'U1_PHASE': 0.18,
        'U2_PHASE': 0.18
            }

    for i in range(len(phases)-1):
        model.setInterfacialEnergy(gamma[phases[i+1]], phase=phases[i+1])
        model.setVolumeBeta(1e-5, VolumeParameter.MOLAR_VOLUME, 4, phase=phases[i+1])
        model.setThermodynamics(AlMgSitherm, phase=phases[i+1])

    model.setup()
    t, x = model.getCurrentX()
    origLen = 5

    x_flat = model.flattenX(x)
    flatShape = x_flat.shape

    x_restore = model.unflattenX(x_flat, x)

    assert(len(x) == origLen)
    assert(np.all(psd.shape == (bins,) for psd in x))
    assert(flatShape == (origLen*bins,))
    assert(len(x_restore) == origLen)
    assert(np.all(psd.shape == (bins,) for psd in x_restore))


from .KWNBase import PrecipitateBase, VolumeParameter
from .KWNEuler import PrecipitateModel
from .PopulationBalance import PopulationBalanceModel
from .non_ideal.ElasticFactors import StrainEnergy
from .non_ideal.ShapeFactors import ShapeFactor
from .TimeTemperaturePrecipitation import TTPCalculator


from kawin.solver.Solver import SolverType, DESolver
import numpy as np
from typing import List
import copy

class GenericModel:
    '''
    Abstract model that new models can inherit from to interface with the Solver
    
    The model is intended to be defined by an ordinary differential equation or a set of them
    The differential equations are defined by dX/dt = f(t,X)
        Where t is time and X is the set of time-dependent variables at time t

    Required functions to be implemented:
        getCurrentX(self)              - should return time and all time-dependent variables
        getdXdt(self, t, x)            - should return all time-dependent derivatives
        getDt(self, dXdt)              - should return a suitable time step
        

    Functions that can be implemented but not necessary:
        _getVarDict(self) - returns a dictionary of {variable name : member name}
        _addExtraSaveVariables(self, saveDict) - adds to saveDict additional variables to save
        _loadExtraVariables(self, data) - loads additional data to model

        setup(self) - ran before solver is called
        correctdXdt(self, dt, x, dXdt) - does not need to return anything, but should modify dXdt
        preProcess(self) - preprocessing before each iteration
        postProcess(self, time, x) - postprocessing after each iteration
        printHeader(self) - initial output statements before solver is called
        printStatus(self, iteration, modelTime, simTimeElapsed) - output states made after n iterations
    '''
    def __init__(self):
        self.clearCouplingModels()

    def _getVarDict(self):
        '''
        Returns variable dictionary mapping variable name to internal member name

        This is used to when saving the model into a npz format, where the member names
        will be replaced with the variable names defined by this dictionary
        '''
        return {}
    
    def _addExtraSaveVariables(self, saveDict):
        '''
        Adds extra variables to the save dictionary that are not covered by the variable dictionary
            The variable dictionary only cover members that can be retrieved from getattr, so
            this function is used to save data if it is from another class that itself is an attribute
        
        Parameters
        ----------
        saveDict : dictionary { str : np.ndarray }
            Dictionary to add data to
        '''
        return
    
    def _loadExtraVariables(self, data):
        '''
        Loads extra variables in data not covered by the variable dictionary

        Parameters
        ----------
        data : dictionary { str : np.ndarray }
            Dictionary to read data from
        '''
        return
    
    def save(self, filename, compressed = True):
        '''
        Saves model data into file

        1. Store model attributes into saveDict using mapping defined from _getVarDict
        2. Add extra variables to saveDict if needed
        3. Save data into .npz format

        Parameters
        ----------
        filename : str
            File name to save to
        compressed : bool (defaults to True)
            Whether to save in compressed format
        '''
        varDict = self._getVarDict()
        saveDict = {}
        for var in varDict:
            saveDict[var] = getattr(self, varDict[var])
        self._addExtraSaveVariables(saveDict)
        if compressed:
            np.savez_compressed(filename, **saveDict)
        else:
            np.savez(filename, **saveDict)

    def _loadData(self, data):
        '''
        Loads data taken from .npz file into model

        1. Sets attributes using mapping defined from _getVarDict
        2. Loads extra variables using _loadExtraVariables

        Parameters
        ----------
        data : dictionary { str : np.ndarray }
            Data to load from
        '''
        varDict = self._getVarDict()
        for var in varDict:
            setattr(self, varDict[var], data[var])
        self._loadExtraVariables(data)

    def addCouplingModel(self, model):
        '''
        Adds a coupling model to the KWN model

        These will be updated after each iteration with the new values of the model

        Parameters
        ----------
        model : object
            Must have a function called updateCoupledModel that takes in a KWNBase or KWNEuler object
        '''
        self.couplingModels.append(model)

    def clearCouplingModels(self):
        '''
        Clears list of coupling models

        Note - this will not reset the coupling models, just removes them from the list
        '''
        self.couplingModels = []

    def updateCoupledModels(self):
        '''
        Updates coupled models with current values
        '''
        for cm in self.couplingModels:
            cm.updateCoupledModel(self)

    def setup(self):
        '''
        Sets up model before being solved

        This is the first thing that is called when the solve function is called

        Note: this will be called each time the solve function called, so if setup only needs to
              be called once, then make sure there's a check in the model implementation to prevent
              setup from being called more than once
        '''
        pass

    def getCurrentX(self):
        '''
        Gets values of time-dependent variables at current time

        The required format of X is not strict as long as it matches dXdt
        Example: if X is a nested list of [[a, b], c], then dXdt should be [[da/dt, db/dt], dc/dt]

        Note: X should only be for variables that are solved by dX/dt = f(t,X)
        Variables that can be computed directly from X should be calculated in the preProcess or postProcess functions

        Returns
        -------
        t : current time of model
        X : unformatted list of floats
        '''
        raise NotImplementedError()

    def getDt(self, dXdt):
        '''
        Gets suitable time step based off dXdt

        Parameters
        ----------
        dXdt : unformated list of floats
            Time derivatives that may be used to find dt

        Returns
        -------
        dt : float
        '''
        raise NotImplementedError()
    
    def getdXdt(self, t, x):
        '''
        Gets dXdt from current time and X

        Parameters
        ----------
        t : float
            Current time
        x : unformated list of floats
            Current values of time-dependent variables
        
        Returns
        -------
        dXdt : unformated list of floats
            Must be in same format as x
        '''
        raise NotImplementedError()

    def correctdXdt(self, dt, x, dXdt):
        '''
        Intended for cases where dXdt can only be corrected once dt is known
            For example, the time derivatives in the population balance model in PrecipitateModel needs to be
                adjusted to avoid negative bins, but this can only be done once dt is known

        If dXdt can be corrected without knowing dt, then it is recommended to be done during the getdXdt function

        No return value, dXdt is to be modified directly
        '''
        pass
    
    def preProcess(self):
        '''
        Performs any pre-processing before an iteration. This may include some calculations or storing temporary variables
        '''
        pass
    
    def postProcess(self, time, x):
        '''
        Post processing done after an iteration

        This should at least involve storing the new values of time and X
        But this can also include additional calculations or return a signal to stop simulations

        Parameters
        ----------
        time : float
            New time
        x : unformatted list of floats
            New values of X

        Returns
        -------
        x : unformatted list of floats
            This is in case X was modified in postProcess
        stop : bool
            If the simulation needs to end early (ex. a stopping condition is met), then return True to stop solving
        '''
        return x, False

    def printHeader(self):
        '''
        First output to be printed when solve is called

        verbose must be True when calling solve
        '''
        print('Iteration\tSim Time(s)\tRun Time(s)')

    def printStatus(self, iteration, modelTime, simTimeElapsed):
        '''
        Output to be printed after n iterations (defined by vIt in solve)

        verbose must be True when calling solve
        '''
        print('{}\t\t{:.1e}\t\t{:.1f}'.format(iteration, modelTime, simTimeElapsed))

    def setTimeInfo(self, currTime, simTime):
        '''
        Store time variables for starting, final and delta time

        This is sometimes useful for determining the time step
        '''
        self.deltaTime = simTime
        self.startTime = currTime
        self.finalTime = currTime+simTime

    def flattenX(self, X):
        '''
        Since X can be a nested list of values or arrays (or anything),
        we want some instructions for the solver and Iterator for how to convert X
        to a 1D array

        By default, we'll assume X is a list of either floats or 1D arrays

        For more complex nesting, this function should be overloaded

        Parameters
        ----------
        X : list of arrays

        Returns
        -------
        X_flat : 1D numpy array
        '''
        return np.hstack(X)
    
    def unflattenX(self, X_flat, X_ref):
        '''
        Converts flattened X array to original nested X

        Parameters
        ----------
        X_flat : 1D numpy array
            Flattened array
        X_ref : list of arrays
            Template to convert X_flat to

        Returns
        -------
        X_new : unflattened list in the same format as X_ref
        '''
        #Not sure if this is the most efficient way, but we can't assume how the nested list in X_ref is structured
        #This should be a shallow copy though, so maybe it's fine
        X_new = copy.copy(X_ref)
        n = 0
        for i in range(len(X_new)):
            #We can't be sure that X_new[i] a python scalar or numpy scalar, so we'll convert to an np.ndarray first
            if len(np.array(X_new[i]).shape) == 0:
                X_new[i] = X_flat[n]
                n += 1
            else:
                arrLen = np.prod(np.array(X_new[i]).shape)
                X_new[i] = np.reshape(X_flat[n:n+arrLen], np.array(X_new[i]).shape)
                n += arrLen
        return X_new

    def solve(self, simTime, solverType = SolverType.RK4, verbose=False, vIt=10, minDtFrac = 1e-8, maxDtFrac = 1):
        '''
        Solves model using the DESolver

        Steps:
            1. Call setup
            2. Create DESolver object and set necessary functions
            3. Get current values of t and X
            4. Solve from current t to t+simTime

        Parameters
        ----------
        simTime : float
            Simulation time (as a delta from current time)
        solverType : SolverType or Iterator (defaults to SolverType.RK4)
            Defines what iteration scheme to use
        verbose : bool (defaults to False)
            Outputs status if true
        vIt : integer (defaults to 10)
            Number of iterations before printing status
        minDtFrac : float (defaults to 1e-8)
            Minimum dt as fraction of simulation time
        maxDtFrac : float (defaults to 1)
            Maximum dt as fraction of simulation time
        '''
        self.setup()

        solver = DESolver(solverType, minDtFrac = minDtFrac, maxDtFrac = maxDtFrac)
        solver.setFunctions(preProcess=self.preProcess, postProcess=self.postProcess, printHeader=self.printHeader, printStatus=self.printStatus)
        solver.setdXdtFunctions(self.getdXdt, self.correctdXdt, self.getDt, self.flattenX, self.unflattenX)
        
        t, X0 = self.getCurrentX()
        self.setTimeInfo(t, simTime)
        solver.solve(self.startTime, X0, self.finalTime, verbose, vIt)
        #solver.solve(self.getdXdt, self.startTime, X0, self.finalTime, verbose, vIt, self.correctdXdt, self.flattenX, self.unflattenX)

class Coupler(GenericModel):
    '''
    Class for coupling multiple GenericModel objects together

    Note:
        coupleddXdt, coupledPreProcess and coupledPostProcess aren't really necessary since
        tighter coupling can also be done by overloading the getdXdt, preProcess and/or postProcess
        functions and calling the method of the Coupler before anything else
        Ex. tighter coupling can be done by
            a)  Overloading coupleddXdt as
                    def coupleddXdt(self, dXdt):
                        ===
                        Modify dXdt here
                        ===
            
            b)  Overriding getdXdt as
                    def getdXdt(self, t, x):
                        dXdt = super().getdXdt(t, x)
                        ---
                        modify dXdt here
                        ---
                        return dXdt

    Parameters
    ----------
    models : List[GenericModel]
        List of models to be solved
    '''
    def __init__(self, models : List[GenericModel]):
        self.models = models

        #Internal time to record
        #We have the option to solve a model for a given amount of time before coupling it
        #  to another model, which would make each model have a different internal time
        #  Thus, we'll record time here as well representing the time during the coupling
        self.time = np.zeros(1)

    def setup(self):
        '''
        Sets up each model
        '''
        super().setup()
        for m in self.models:
            m.setup()

    def setTimeInfo(self, currTime, simTime):
        '''
        Sets time info for the CoupledModel class and each model
        '''
        super().setTimeInfo(currTime, simTime)
        for m in self.models:
            m.setTimeInfo(currTime, simTime)

    def flattenX(self, X):
        '''
        Instructions for converting X to 1D array

        We grab the flattened x array of each model and concatenate them
            Thus we don't have to care about the structure of x in each model as
            long as the model itself has the instructions to flatten its x array

        Also record the length of each flattened x in each model so we know what
        indices to used for unflattening
        '''
        X_new = []
        for m, xsub in zip(self.models, X):
            xsub_new = m.flattenX(xsub)
            X_new.append(xsub_new)
        self._sizeRef = [len(xi) for xi in X_new]
        return np.concatenate(X_new)

    def unflattenX(self, X_flat, X_ref):
        '''
        Instructions for converting X_flat to list of x of each model

        We take the subset of X_flat corresponding to each model and unflatten it
        based off the instructions in the model. Then we just return a list containing
        each unflattened x
        '''
        X_new = []
        ind = 0
        for m, s, x_refsub in zip(self.models, self._sizeRef, X_ref):
            xi_new = m.unflattenX(X_flat[ind:ind+s], x_refsub)
            X_new.append(xi_new)
            ind += s
        return X_new

    def getCurrentX(self):
        '''
        Get current time and x for each model
        '''
        xs = []
        for m in self.models:
            _, x = m.getCurrentX()
            xs.append(x)

        return self.time[-1], xs
    
    def getDt(self, dXdt):
        '''
        Get the minimum dt out of all models
        '''
        dts = []
        for m, dxdtsub in zip(self.models, dXdt):
            dts.append(m.getDt(dxdtsub))
        return np.amin(dts)
    
    def getdXdt(self, t, x):
        '''
        Get dXdt for each model
        '''
        dxdts = []
        for m, xsub in zip(self.models, x):
            dxdts.append(m.getdXdt(t, xsub))
        self.coupledXdt(t, x, dxdts)
        return dxdts
    
    def correctdXdt(self, dt, x, dXdt):
        '''
        Corrects dXdt for each model

        Note - dXdt has to be modified since we don't return dXdt in this function
            Since dXdt here is composed of a nested list of dXdts of each model, these
            will be passed by reference
        '''
        for m, xsub, dxdtsub in zip(self.models, x, dXdt):
            m.correctdXdt(dt, xsub, dxdtsub)

    def preProcess(self):
        '''
        Pre process on each model
        '''
        for m in self.models:
            m.preProcess()
        self.couplePreProcess()

    def postProcess(self, time, x):
        '''
        Post process on each model and records new time
        '''
        xNew = []
        stop = False
        for m, xsub in zip(self.models, x):
            xnew_sub, s = m.postProcess(time, xsub)
            stop = stop or s
            xNew.append(xnew_sub)
        self.time = np.append(self.time, time)
        self.couplePostProcess()
        return xNew, stop
    
    def coupledXdt(self, t, x, dXdt):
        '''
        Empty function where inherited classes can do extra operations on
        the time derivatives each models or between models
        '''
        return
    
    def couplePreProcess(self):
        '''
        Empty function where inherited classes can do extra operations on
        each models or between models for an iteration
        '''
        return
    
    def couplePostProcess(self):
        '''
        Empty function where inherited classes can do extra operations on
        each models or between models after an iteration
        '''
        return



        


import numpy as np
from pycalphad import Workspace, Model, Database, calculate, equilibrium, variables as v
from pycalphad.codegen.phase_record_factory import PhaseRecordFactory
from pycalphad.core.composition_set import CompositionSet
from pycalphad.core.utils import extract_parameters
from kawin.thermo.Mobility import MobilityModel, inverseMobility, inverseMobility_from_diffusivity, tracer_diffusivity, tracer_diffusivity_from_diff
from kawin.thermo.FreeEnergyHessian import dMudX
from kawin.thermo.LocalEquilibrium import local_equilibrium
import copy
from tinydb import where

class ExtraFreeEnergyType(v.IndependentPotential):
    implementation_units = 'joules'
    display_units = 'joules'
    display_name = 'Extra Gibbs Free Energy'

    def __init__(self):
        super().__init__('GE')
    def __reduce__(self):
        return self.__class__, ()
    
setattr(v, 'GE', v.IndependentPotential('GE'))

class ExtraGibbsModel(Model):
    '''
    Child of pycalphad Model with extra variable GE
        GE represents any extra contribution to the Gibbs free energy
        such as the Gibbs-Thomson contribution
    '''
    energy = GM = property(lambda self: self.ast + v.GE)
    formulaenergy = G = property(lambda self: (self.ast + v.GE) * self._site_ratio_normalization)
    orderingContribution = OCM = property(lambda self: self.models['ord'])

class GeneralThermodynamics:
    '''
    Class for defining driving force and essential functions for
    binary and multicomponent systems using pycalphad for equilibrium
    calculations

    Parameters
    ----------
    database : Database or str
        pycalphad Database or file name for database
    elements : list
        Elements to consider
        Note: reference element must be the first index in the list
    phases : list
        Phases involved
        Note: matrix phase must be first index in the list
    drivingForceMethod : str (optional)
        Method used to calculate driving force
        Options are 'tangent' (default), 'approximate', 'sampling' and 'curvature' (not recommended)
    parameters : list [str] or dict {str : float} or None
        List of parameters to keep symbolic in the thermodynamic or mobility models
        If None, then parameters are fixed
    '''

    gOffset = 1      #Small value to add to precipitate phase for when order/disorder models are used

    def __init__(self, database, elements, phases, drivingForceMethod = 'tangent', parameters = None):
        if isinstance(database, str):
            database = Database(database)
        self.db = database
        self.elements = copy.copy(elements)
        if parameters is None:
            self._parameters = {}
        else:
            if isinstance(parameters, list):
                self._parameters = {p: 0 for p in parameters}
            else:
                self._parameters = parameters

        if 'VA' not in self.elements:
            self.elements.append('VA')

        if type(phases) == str:  # check if a single phase was passed as a string instead of a list of phases.
            phases = [phases]
        self.phases = phases

        self._buildThermoModels()

        #Amount of points to sample per degree of freedom
        # sampling_pDens is for when using sampling method in driving force calculations
        # pDens is for equilibrium calculations
        self.sampling_pDens = 2000
        self.pDens = 500

        #Stored variables of last time the class was used
        #This is so that these can be used again if the temperature has not changed since last usage
        self._prevTemperature = None

        #Pertains to parent phase (composition, sampled points, equilibrium calculations)
        self._prevX = None
        self._parentEq = None

        #Pertains to precipitate phases (sampled points)
        self._pointsPrec = {self.phases[i]: None for i in range(1, len(self.phases))}
        self._orderingPoints = {self.phases[i]: None for i in range(1, len(self.phases))}

        self.setDrivingForceMethod(drivingForceMethod)

        self._buildMobilityModels()

        #Cached results
        self._compset_cache = {}
        self._compset_cache_df = {}
        self._matrix_cs = None

    def _buildThermoModels(self):
        '''
        Builds thermodynamic models for each phase

        This assumes that the first phase is the parent phase and the rest of the phases are precipitate phases
            For usage in a diffusion model, this won't affect anything

        For each precipitate phase, it checks whether the phase has an order/disorder contribution
            If so, then it checks if the disorder contribution comes from the parent phase (ex. gamma and gamma prime in Ni alloys)
            An ordering contribution phase record will be created to allow separated between the parent and the ordered phase
        '''
        self.orderedPhase = {self.phases[i]: False for i in range(1, len(self.phases))}
        for i in range(1, len(self.phases)):
            if 'disordered_phase' in self.db.phases[self.phases[i]].model_hints:
                if self.db.phases[self.phases[i]].model_hints['disordered_phase'] == self.phases[0]:
                    self.orderedPhase[self.phases[i]] = True
                    self._forceDisorder(self.phases[0])

        #Build phase models assuming first phase is parent phase and rest of precipitate phases
        #If the same phase is used for matrix and precipitate phase, then force the matrix phase to remove the ordering contribution
        #This may be unnecessary as already disordered phase models will not be affected, but I guess just in case the matrix phase happens to be an ordered solution
        param_keys, _ = extract_parameters(self._parameters)
        self.models = {self.phases[0]: Model(self.db, self.elements, self.phases[0], parameters=param_keys)}
        self.models[self.phases[0]].state_variables = sorted([v.T, v.P, v.N, v.GE], key=str)

        for i in range(1, len(self.phases)):
            self.models[self.phases[i]] = ExtraGibbsModel(self.db, self.elements, self.phases[i], parameters=param_keys)
            self.models[self.phases[i]].state_variables = sorted([v.T, v.P, v.N, v.GE], key=str)

        self.phase_records = PhaseRecordFactory(self.db, self.elements, 
                                                       self.models[self.phases[0]].state_variables, 
                                                       self.models, parameters=self._parameters)

    def _buildMobilityModels(self):
        '''
        Builds mobility models for phases that have model parameters
        '''
        self.mobModels = {p: None for p in self.phases}
        self.mobCallables = {p: None for p in self.phases}
        self.diffCallables = {p: None for p in self.phases}
        param_keys, _ = extract_parameters(self._parameters)
        phase_mob_params = {}
        phase_diff_params = {}
        self.mobModels = {}
        for p in self.phases:
            #Get mobility/diffusivity of phase p if exists
            param_search = self.db.search
            param_query_mob = (
                (where('phase_name') == p) & \
                (where('parameter_type') == 'MQ') | \
                (where('parameter_type') == 'MF')
            )

            param_query_diff = (
                (where('phase_name') == p) & \
                (where('parameter_type') == 'DQ') | \
                (where('parameter_type') == 'DF')
            )
            phase_mob_params[p] = param_search(param_query_mob)
            phase_diff_params[p] = param_search(param_query_diff)

            if len(phase_mob_params[p]) > 0 or len(phase_diff_params[p]) > 0:
                self.mobModels[p] = MobilityModel(self.db, self.elements, p, parameters=param_keys)
                self.mobModels[p].state_variables = sorted([v.T, v.P, v.N, v.GE], key=str)

        mob_phases = list(self.mobModels.keys())
        self.mob_phase_records = PhaseRecordFactory(self.db, self.elements, 
                                                       self.mobModels[mob_phases[0]].state_variables, 
                                                       self.mobModels, parameters=self._parameters)

        for p in self.mobModels:
            if len(phase_mob_params[p]) > 0:
                self.mobCallables[p] = {c: self.mob_phase_records.get_phase_property(p, 'MOB_'+c, include_grad=False, include_hess=False).func for c in self.phase_records[p].nonvacant_elements}
            if len(phase_diff_params[p]) > 0:
                self.diffCallables[p] = {c: self.mob_phase_records.get_phase_property(p, 'DIFF_'+c, include_grad=False, include_hess=False).func for c in self.phase_records[p].nonvacant_elements}

        #This applies to all phases since this is typically reflective of quenched-in vacancies
        self.mobility_correction = {A: 1 for A in self.elements}

    def updateParameters(self, parameters):
        '''
        Update parameter dictionary with new values

        Parameters
        ----------
        parameters : dict {str : float}
            Dictionary of parameters
            NOTE: this does not have to be the full list and can also have other parameters in it
                Only the parameters that are stored upon initialization will be changed
        '''
        for pm in parameters:
            if pm in self._parameters:
                self._parameters[pm] = parameters[pm]

        param_keys, param_values = extract_parameters(self._parameters)
        for p in self.phases:
            self.phase_records[p].parameters[:] = np.asarray(param_values, dtype=np.float64)

    def _forceDisorder(self, phase):
        '''
        For phases using an order/disorder model, pycalphad will neglect the disordered phase unless
        it is the only phase set active, so the order and disordered portion of the phase will use the same model

        For the Gibbs-Thomson effect to be applied, the ordered and disordered parts of the model will need to be kept separate
        As a fix, a new phase is added to the database that uses only the disordered part of the model
        '''
        newPhase = 'DIS_' + phase
        self.phases[0] = newPhase
        self.db.phases[newPhase] = copy.deepcopy(self.db.phases[phase])
        self.db.phases[newPhase].name = newPhase
        del self.db.phases[newPhase].model_hints['ordered_phase']
        del self.db.phases[newPhase].model_hints['disordered_phase']

        #Copy database parameters with new name
        param_query = where('phase_name') == phase
        params = self.db.search(param_query)
        for p in params:
            #We have to create a new dictionary since p is a TinyDB.Document
            newP = {}
            for entry in p:
                newP[entry] = p[entry]
            newP['phase_name'] = newPhase
            self.db._parameters.insert(newP)

    def clearCache(self):
        '''
        Removes any cached data
        This is intended for surrogate training, where the cached data
        will be removed incase
        '''
        self._compset_cache = {}
        self._compset_cache_df = {}
        self._matrix_cs = None

    def setDrivingForceMethod(self, drivingForceMethod):
        '''
        Sets method for calculating driving force

        Parameters
        ----------
        drivingForceMethod - str
            Options are ['approximate', 'sampling', 'curvature']
        '''
        if drivingForceMethod == 'approximate':
            self._drivingForce = self._getDrivingForceApprox
        elif drivingForceMethod == 'sampling':
            self._drivingForce = self._getDrivingForceSampling
        elif drivingForceMethod == 'curvature':
            self._drivingForce = self._getDrivingForceCurvature
        elif drivingForceMethod == 'tangent':
            self._drivingForce = self._getDrivingForceTangent
        else:
            raise Exception('Driving force method must be either \'approximate\', \'sampling\', \'tangent\' or \'curvature\'')

    def setDFSamplingDensity(self, density):
        '''
        Sets sampling density for sampling method in driving
        force calculations

        Default upon initialization is 2000

        Parameters
        ----------
        density : int
            Number of samples to take per degree of freedom in the phase
        '''
        self._pointsPrec = {self.phases[i]: None for i in range(1, len(self.phases))}
        self.sampling_pDens = density

    def setEQSamplingDensity(self, density):
        '''
        Sets sampling density for equilibrium calculations

        Default upon initialization is 500

        Parameters
        ----------
        density : int
            Number of samples to take per degree of freedom in the phase
        '''
        self.pDens = density

    def setMobility(self, mobility):
        '''
        Allows user to define mobility functions

        mobility : dict
            Dictionary of functions for each element (including reference)
            Each function takes in (v.T, v.P, v.N, v.GE, site fractions) and returns mobility

        Optional - only required for multicomponent systems where
            mobility terms are not defined in the TDB database
        '''
        self.mobCallables = mobility

    def setDiffusivity(self, diffusivity):
        '''
        Allows user to define diffusivity functions

        diffusivity : dict
            Dictionary of functions for each element (including reference)
            Each function takes in (v.T, v.P, v.N, v.GE, site fractions) and returns diffusivity

        Optional - only required for multicomponent systems where
            diffusivity terms are not defined in the TDB database
            and if mobility terms are not defined
        '''
        self.diffCallables = diffusivity

    def setMobilityCorrection(self, element, factor):
        '''
        Factor to multiply mobility by for each element

        Parameters
        ----------
        element : str
            Element to set factor for
            If 'all', factor will be set to all elements
        factor : float
            Scaling factor
        '''
        if element == 'all':
            for e in self.mobility_correction:
                self.mobility_correction[e] = factor
        else:
            self.mobility_correction[element] = factor

    def _getConditions(self, x, T, gExtra = 0):
        '''
        Creates dictionary of conditions from composition, temperature and gExtra

        Parameters
        ----------
        x : list
            Composition (excluding reference element)
        T : float
            Temperature
        gExtra : float
            Gibbs free energy to add to phase
        '''
        cond = {v.X(self.elements[i+1]): x[i] for i in range(len(x))}
        cond[v.P] = 101325
        cond[v.T] = T
        cond[v.GE] = gExtra
        cond[v.N] = 1
        return cond

    def _createCompositionSet(self, eq, state_variables, phase, phase_amounts, idx):
        '''
        Creates a pycalphad CompositionSet from equilibrium results

        Parameters
        ----------
        eq : pycalphad equilibrium result
        state_variables : list
            List of state variables
        phase : str
            Phase to create CompositionSet for
        phase_amounts : list
            Array of floats for phase fraction of each phase
        idx : ndarray
            Index array for the index of phase
        '''
        miscibility = False
        cs = CompositionSet(self.phase_records[phase])
        #If there's a miscibility gap in the matrix phase, then take the largest value
        if len(idx) > 1:
            idx = [idx[np.argmax(phase_amounts[idx])]]
            miscibility = True
        cs.update(eq.Y.isel(vertex=idx[0]).values.ravel()[:cs.phase_record.phase_dof],
                        phase_amounts[idx[0]], state_variables)

        return cs, miscibility
    
    def _setupSubModels(self, precPhase = None):
        phases = [self.phases[0]]
        if precPhase != -1:
            if precPhase is None:
                precPhase = self.phases[1]
            if isinstance(precPhase, str):
                phases.append(precPhase)
            else:
                phases = [p for p in precPhase]

        sub_models = {p: self.models[p] for p in phases}
        self.phase_records.models = sub_models
        return phases, sub_models

    def getEq(self, x, T, gExtra = 0, precPhase = None):
        '''
        Calculates equilibrium at specified x, T, gExtra

        This is separated from the interfacial composition function so that this can be used for getting curvature for interfacial composition from mobility

        Parameters
        ----------
        x : float or array
            Composition
            Needs to be array for multicomponent systems
        T : float
            Temperature
        gExtra : float
            Gibbs-Thomson contribution (if applicable)
        precPhase : str, int, list or None
            Precipitate phase (default is first precipitate)
            Options:
                None - first precipitate phase in phase list
                str - specific precipitate phase by name
                list - all phases by name in list
                -1 - no precipitate phase

        Returns
        -------
        Dataset from pycalphad equilibrium results
        '''
        if not hasattr(x, '__len__'):
            x = [x]

        #Remove first element if x lists composition of all elements
        if len(x) == len(self.elements) - 1:
            x = x[1:]

        cond = self._getConditions(x, T, gExtra+self.gOffset)
        
        # Phases, models and phase records should match in the list of phases
        # Here, I create a model dictionary of the relevant phases
        # Then, I replace the models in the phase records with this sublist
        # After equilibrium, restore the original model list
        phases, sub_models = self._setupSubModels(precPhase)
        sub_models = {p: self.models[p] for p in phases}
        wks = Workspace(self.db, self.elements, phases, cond, models=sub_models, phase_record_factory=self.phase_records, calc_opts={'pdens': self.pDens})

        return wks

    def getLocalEq(self, x, T, gExtra = 0, precPhase = None, composition_sets = None):
        '''
        Calculates local equilibrium at specified x, T, gExtra

        Parameters
        ----------
        x : float or array
            Composition
            Needs to be array for multicomponent systems
        T : float
            Temperature
        gExtra : float
            Gibbs-Thomson contribution (if applicable)
        precPhase : str, int, list or None
            Precipitate phase (default is first precipitate)
            Options:
                None - first precipitate phase in phase list
                str - specific precipitate phase by name
                list - all phases by name in list
                -1 - no precipitate phase

        Returns
        -------
        result - equilibrium convergence and chemical potentials
        composition_sets - list of CompositionSet for phases in "equilibrium"
            Note - "equilibrium" in terms of the matrix and singled out precipitate phase (or just matrix if precPhase is -1)
        '''
        if not hasattr(x, '__len__'):
            x = [x]

        #Remove first element if x lists composition of all elements
        if len(x) == len(self.elements) - 1:
            x = x[1:]

        cond = self._getConditions(x, T, gExtra)
        phases, sub_models = self._setupSubModels(precPhase)
        result, composition_sets = local_equilibrium(self.db, self.elements, phases, cond,
                                                         sub_models, self.phase_records,
                                                         composition_sets=composition_sets)
        return result, composition_sets

    def getInterdiffusivity(self, x, T, removeCache = True, phase = None):
        '''
        Gets interdiffusivity at specified x and T
        Requires TDB database to have mobility or diffusivity parameters

        Parameters
        ----------
        x : float, array or 2D array
            Composition
            Float or array for binary systems
            Array or 2D array for multicomponent systems
        T : float or array
            Temperature
            If array, must be same length as x
                For multicomponent systems, must be same length as 0th axis
        removeCache : boolean
            If True, recalculates equilibrium to get interdiffusivity (default)
            If False, will use calculation from driving force calcs (if available) to compute diffusivity
        phase : str
            Phase to compute diffusivity for (defaults to first or matrix phase)
            This only needs to be used for multiphase diffusion simulations

        Returns
        -------
        interdiffusivity - will return array if T is an array
            For binary case - float or array of floats
            For multicomponent - matrix or array of matrices
        '''
        dnkj = []

        if hasattr(T, '__len__'):
            for i in range(len(T)):
                dnkj.append(self._interdiffusivitySingle(x[i], T[i], removeCache, phase))
            return np.array(dnkj)
        else:
            return self._interdiffusivitySingle(x, T, removeCache, phase)

    def _interdiffusivitySingle(self, x, T, removeCache = True, phase = None):
        '''
        Gets interdiffusivity at unique composition and temperature

        Parameters
        ----------
        x : float or array
            Composition
        T : float
            Temperature
        removeCache : boolean
        phase : str

        Returns
        -------
        Interdiffusivity as a matrix (will return float in binary case)
        '''
        if phase is None:
            phase = self.phases[0]

        if not hasattr(x, '__len__'):
            x = [x]

        #Remove first element if x lists composition of all elements
        if len(x) == len(self.elements) - 1:
            x = x[1:]

        cond = self._getConditions(x, T, 0)

        if removeCache:
            self._matrix_cs = None
            phases, sub_models = self._setupSubModels([phase])
            self._parentEq, self._matrix_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                        sub_models, self.phase_records,
                                                        composition_sets=self._matrix_cs)

        cs_matrix = [cs for cs in self._matrix_cs if cs.phase_record.phase_name == phase][0]
        chemical_potentials = self._parentEq.chemical_potentials

        if self.mobCallables[phase] is None:
            Dnkj, _, _ = inverseMobility_from_diffusivity(chemical_potentials, cs_matrix,
                                                                            self.elements[0], self.diffCallables[phase],
                                                                            diffusivity_correction=self.mobility_correction,
                                                                            parameters = self._parameters)
        else:
            Dnkj, _, _ = inverseMobility(chemical_potentials, cs_matrix, self.elements[0],
                                                        self.mobCallables[phase],
                                                        mobility_correction=self.mobility_correction,
                                                        parameters=self._parameters)

        if len(x) == 1:
            return Dnkj.ravel()[0]
        else:
            sortIndices = np.argsort(self.elements[1:-1])
            unsortIndices = np.argsort(sortIndices)
            Dnkj = Dnkj[unsortIndices,:]
            Dnkj = Dnkj[:,unsortIndices]
            return Dnkj


    def getTracerDiffusivity(self, x, T, removeCache = True, phase = None):
        '''
        Gets tracer diffusivity for element el at specified x and T
        Requires TDB database to have mobility or diffusivity parameters

        Parameters
        ----------
        x : float, array or 2D array
            Composition
            Float or array for binary systems
            Array or 2D array for multicomponent systems
        T : float or array
            Temperature
            If array, must be same length as x
                For multicomponent systems, must be same length as 0th axis
        removeCache : boolean
        phase : str

        Returns
        -------
        tracer diffusivity - will return array if T is an array
        '''
        td = []

        if hasattr(T, '__len__'):
            for i in range(len(T)):
                td.append(self._tracerDiffusivitySingle(x[i], T[i], removeCache, phase))
            return np.array(td)
        else:
            return self._tracerDiffusivitySingle(x, T, removeCache, phase)

    def _tracerDiffusivitySingle(self, x, T, removeCache = True, phase = None):
        '''
        Gets tracer diffusivity at unique composition and temperature

        Parameters
        ----------
        x : float or array
            Composition
        T : float
            Temperature
        el : str
            Element to calculate diffusivity
        
        Returns
        -------
        Tracer diffusivity as a float
        '''
        if phase is None:
            phase = self.phases[0]

        if not hasattr(x, '__len__'):
            x = [x]

        #Remove first element if x lists composition of all elements
        if len(x) == len(self.elements) - 1:
            x = x[1:]

        cond = self._getConditions(x, T, 0)

        if removeCache:
            self._matrix_cs = None
            phases, sub_models = self._setupSubModels([phase])
            self._parentEq, self._matrix_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                        sub_models, self.phase_records,
                                                        composition_sets=self._matrix_cs)

        cs_matrix = [cs for cs in self._matrix_cs if cs.phase_record.phase_name == phase][0]

        if self.mobCallables[phase] is None:
            #NOTE: This is not tested yet
            Dtrace = tracer_diffusivity_from_diff(cs_matrix, self.diffCallables[phase], diffusivity_correction=self.mobility_correction, parameters=self._parameters)
        else:
            Dtrace = tracer_diffusivity(cs_matrix, self.mobCallables[phase], mobility_correction=self.mobility_correction, parameters=self._parameters)

        sortIndices = np.argsort(self.elements[:-1])
        unsortIndices = np.argsort(sortIndices)

        Dtrace = Dtrace[unsortIndices]

        return Dtrace

    def getDrivingForce(self, x, T, precPhase = None, returnComp = False, training = False):
        '''
        Gets driving force using method defined upon initialization

        Parameters
        ----------
        x : float, array or 2D array
            Composition of minor element in bulk matrix phase
            For binary system, use an array for multiple compositions
            For multicomponent systems, use a 2D array for multiple compositions
                Where 0th axis is for indices of each composition
        T : float or array
            Temperature in K
            Must be same length as x if x is array or 2D array
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        (driving force, precipitate composition)
        Driving force is positive if precipitate can form
        Precipitate composition will be None if driving force is negative or returnComp is False
        '''
        if hasattr(T, '__len__'):
            dgArray = []
            compArray = []
            for i in range(len(T)):
                dg, comp = self._drivingForce(x[i], T[i], precPhase, returnComp, training)
                dgArray.append(dg)
                compArray.append(comp)
            dgArray = np.array(dgArray)
            compArray = np.array(compArray)
            return dgArray, compArray
        else:
            return self._drivingForce(x, T, precPhase, returnComp, training)

    def _getDrivingForceSampling(self, x, T, precPhase = None, returnComp = False, training = False):
        '''
        Gets driving force for nucleation by sampling

        Steps
            1. Compute local equilibrium at x and T of only the matrix phase
            2. Sample precipitate phase
                If ordered contribution to matrix phase, then sample ordering contribution
                and remove points on the matrix free energy surface
            3. Compute energy difference between precipitate samples and chemical potential hyperplane
            4. Find sample that maximizes energy difference and return sample composition and driving force

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        (driving force, precipitate composition)
        Driving force is positive if precipitate can form
        Precipitate composition will be None if driving force is negative or returnComp is False
        '''
        precPhase = self.phases[1] if precPhase is None else precPhase

        #Calculate equilibrium with only the parent phase -------------------------------------------------------------------------------------------
        if not hasattr(x, '__len__'):
            x = [x]
        cond = self._getConditions(x, T, 0)
        self._prevX = x

        cs_results = self._getPrecCompositionSetSamplingDF(x, T, cond, precPhase, training)
        if cs_results is None:
            return None, None
        
        dg, prec_cs = cs_results

        #Remove cache when training
        if training:
            self._matrix_cs = None

        if returnComp:
            sortIndices = np.argsort(self.elements[:-1])
            unsortIndices = np.argsort(sortIndices)
            beta_x = np.array(prec_cs.X, dtype=np.float64)
            beta_x = beta_x[unsortIndices]
            if len(x) == 1:
                return dg, beta_x[1:][0]
            else:
                return dg, beta_x[1:]
        else:
            return dg, None

    def _getDrivingForceApprox(self, x, T, precPhase = None, returnComp = False, training = False):
        '''
        Approximate method of driving force calculation
        Assumes equilibrium composition of precipitate phase

        Sampling method is used if driving force is negative

        Steps:
            1. Compute equilibrium and get composition sets for matrix and precipitate phase
            2. Check for 2 phases and that one phase is the matrix and other phase is precipitate
                If not, then resort to sampling method
            3. Compute equilibrium at matrix composition and get chemical potential hyperplane
            4. Driving force is the difference between the free energy of the precipitate (from step 1)
               and the free energy on the chemical potential hyperplane (from step 3) at the precipitate composition

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        (driving force, precipitate composition)
        Driving force is positive if precipitate can form
        Precipitate composition will be None if driving force is negative or returnComp is False
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        if not hasattr(x, '__len__'):
            x = [x]
        cond = self._getConditions(x, T, 0)
        self._prevX = x

        cs_results = self._getCompositionSetsForDF(x, T, cond, precPhase, training=training)
        if cs_results is None:
            return self._getDrivingForceSampling(x, T, precPhase, returnComp, training=training)
        
        chemical_potentials, cs_matrix, cs_precip = cs_results
        sortedElements = sorted(list(set(self.elements) - {'VA'}))
        refIndex = sortedElements.index(self.elements[0])

        #Equilibrium at matrix composition for only the parent phase
        phases, sub_models = self._setupSubModels([self.phases[0]])
        self._parentEq, self._matrix_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                            sub_models, self.phase_records,
                                                            composition_sets=self._matrix_cs)
        if training:
            self._matrix_cs = None

        #Check if equilibrium has converged and chemical potential can be obtained
        #If not, then return None for driving force
        if any(np.isnan(self._parentEq.chemical_potentials)):
            return None, None

        sortIndices = np.argsort(self.elements[1:-1])
        unsortIndices = np.argsort(sortIndices)

        xP = np.array(cs_precip.X, dtype=np.float64)

        dg = np.sum(xP * self._parentEq.chemical_potentials) - np.sum(xP * chemical_potentials)

        #Remove reference element
        xP = np.delete(xP, refIndex)

        if returnComp:
            if len(x) == 1:
                return dg.ravel()[0], xP[unsortIndices][0]
            else:
                return dg.ravel()[0], xP[unsortIndices]
        else:
            return dg.ravel()[0], None

    def _getDrivingForceCurvature(self, x, T, precPhase = None, returnComp = False, training = False):
        '''
        Gets driving force from curvature of free energy function
        Assumes small saturation

        Steps:
            1. Compute equilibrium and get composition sets for matrix and precipitate phase
            2. Check for 2 phases and that one phase is the matrix and other phase is precipitate
                If not, then resort to sampling method
            3. Get dmu/dx (free energy curvature)
            4. Compute (x_infty - x_matrix) * dmu/dx * (x_prec - x_matrix)^T
                This does a first (or second?) order approximation of the driving force based off the curvature at x_infty

        Sampling method is used if driving force is negative

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        (driving force, precipitate composition)
        Driving force is positive if precipitate can form
        Precipitate composition will be None if driving force is negative or returnComp is False
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        if not hasattr(x, '__len__'):
            x = [x]
        cond = self._getConditions(x, T, 0)
        self._prevX = x

        cs_results = self._getCompositionSetsForDF(x, T, cond, precPhase, training=training)
        if cs_results is None:
            return self._getDrivingForceSampling(x, T, precPhase, returnComp, training=training)
        
        chemical_potentials, cs_matrix, cs_precip = cs_results
        sortedElements = sorted(list(set(self.elements) - {'VA'}))
        refIndex = sortedElements.index(self.elements[0])
        x_matrix = np.array(cs_matrix.X, dtype=np.float64)
        x_precip = np.array(cs_precip.X, dtype=np.float64)

        #If in two phase region, then get curvature of parent phase and use it to calculate driving force ---------------------------------------
        sortIndices = np.argsort(self.elements[1:-1])
        unsortIndices = np.argsort(sortIndices)

        dMudxParent = dMudX(chemical_potentials, cs_matrix, self.elements[0])
        xM = np.delete(x_matrix, refIndex)
        xP = np.delete(x_precip, refIndex)
        xBar = np.array([xP - xM])

        x = np.array(x)[sortIndices]
        xD = np.array([x - xM])

        dg = np.matmul(xD, np.matmul(dMudxParent, xBar.T))

        if returnComp:
            if len(x) == 1:
                return dg.ravel()[0], xP[unsortIndices][0]
            else:
                return dg.ravel()[0], xP[unsortIndices]
        else:
            return dg.ravel()[0], None

    def _getDrivingForceTangent(self, x, T, precPhase = None, returnComp = False, training = False):
        '''
        Gets driving force from parallel tangent calculation

        Steps
            1. Compute equilibrium to get composition sets (or used previous cached CS)
            2. Compute equilibrium of matrix phase at matrix composition
            3. Remove composition and extra free energy from conditions
            4. Add chemical potential for each component to conditions
            5. Compute equilibrium of precipitate phase with new conditions
                The calculated v.GE is the driving force

        This will work for positive and negative driving forces

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other
            
        Returns
        -------
        (driving force, precipitate composition)
        Driving force is positive if precipitate can form
        Precipitate composition will be None if driving force is negative or returnComp is False
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        if not hasattr(x, '__len__'):
            x = [x]
        cond = self._getConditions(x, T, self.gOffset)
        self._prevX = x

        if self._compset_cache_df.get(precPhase, None) is None or training:
            #This will calculate local equilibrium for the matrix phase and get the composition set for the precipitate phase
            # _parentEq is set in this function
            cs_results = self._getPrecCompositionSetSamplingDF(x, T, cond, precPhase, training=training)
            if cs_results is None:
                return None, None

            dg, _prec_cs = cs_results
            self._compset_cache_df[precPhase] = [_prec_cs]
        else:
            #If we already have a cache, then we just need equilibrium at the matrix phase
            phases, sub_models = self._setupSubModels([self.phases[0]])
            self._parentEq, self._matrix_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                    sub_models, self.phase_records, composition_sets=self._matrix_cs)
        
            #Check that equilibrium has converged
            #If not, then return None, None since driving force can't be obtained
            if any(np.isnan(self._parentEq.chemical_potentials)):
                return None, None

        #Remove element conditions and free extra Gibbs energy conditions
        for e in self.elements:
            if v.X(e) in cond:
                cond.pop(v.X(e))
        if v.GE in cond:
            cond.pop(v.GE)

        #Add chemical potential conditions
        sortedEl = sorted(list(set(self.elements) - set(['VA'])))
        for i in range(len(sortedEl)):
            cond[v.MU(sortedEl[i])] = self._parentEq.chemical_potentials[i]

        #Solving for local equilibrium on precipitate
        #The fixed conditions are T, P and MU, so this should solve for precipitate composition and GE
        #   Rather than solving for parallel tangent where the driving force is the difference between the chemical potentials of matrix and precipitate phase
        #   This instead solves for the offset in the precipitate energy surface to make the precipitate lie on the chemical potential hyperplane of the matrix phase
        prev_dof = np.array(self._compset_cache_df[precPhase][0].dof)
        phases, sub_models = self._setupSubModels([precPhase])
        _precEq, _prec_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                sub_models, self.phase_records, composition_sets=self._compset_cache_df[precPhase])

        #Check if precipitate composition at equilibrium is the matrix composition
        #This can occur in order/disordered models where the miscibility gap is small enough that the parallel tangent can only be found at the matrix composition
        #In this case, switch to sampling for the driving force
        #This still seems to be an improvement over approximate and curvature methods since this occurs after the driving force becomes negative
        prec_comps = np.array(_prec_cs[0].X)
        mat_comps = np.array(self._matrix_cs[0].X)
        if np.allclose(prec_comps, mat_comps, 1e-6):
            self._compset_cache_df[precPhase] = None
            return self._getDrivingForceSampling(x, T, precPhase, returnComp, training=training)

        self._compset_cache_df[precPhase] = _prec_cs

        #Check that equilibrium has converged
        #If not, then return None, None since driving force can't be obtained
        if any(np.isnan(_precEq.chemical_potentials)):
            return None, None
        
        dg = _precEq.x[0]
        xb = np.array(_prec_cs[0].X)

        sortIndices = np.argsort(self.elements[:-1])
        unsortIndices = np.argsort(sortIndices)
        xb = xb[unsortIndices]

        if returnComp:
            if len(x) == 1:
                return dg, xb[1:][0]
            else:
                return dg, xb[1:]
        else:
            return dg, None
    
    def _getCompositionSetsForDF(self, x, T, cond, precPhase = None, training = False):
        '''
        Wrapper for getting composition set from x and T by either global equilibrium or local from a cached composition set

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        chemical_potentials
        cs_matrix - composition set of matrix phase
        cs_precip - composition set of precipitate phase
        '''
        if self._compset_cache_df.get(precPhase, None) is None or training:
            return self._getCompositionSetsEq(x, T, cond, precPhase)
        else:
            return self._getCompositionSetsCache(x, T, cond, precPhase)
    
    def _getCompositionSetsEq(self, x, T, cond, precPhase = None):
        '''
        Gets composition set from x and T by global equilibrium

        Steps
            1. Compute equilibrium at x and T
                If equilibrium did not converge or matrix phase is not stable, then return None
            2. Get composition sets and add to cache
                If precipitate is not stable, the return None
            3. Resolve possible issues with miscibility gaps
            4. Return values

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)

        Returns
        -------
        chemical_potentials
        cs_matrix - composition set of matrix phase
        cs_precip - composition set of precipitate phase
        '''
        #Create cache of composition set if not done so already or if training a surrogate
        #Training points for surrogates may be far apart, so starting from a previous
        #   composition set could give a bad starting position for the minimizer
        #Calculate equilibrium ----------------------------------------------------------------------------------------------------------------------
        wks = self.getEq(x, T, 0, precPhase)
        cs_list_matrix = []
        cs_list_precip = []
        for cs in wks.get_composition_sets():
            if cs.phase_record.phase_name == self.phases[0]:
                cs_list_matrix.append(cs)
            elif cs.phase_record.phase_name == precPhase:
                cs_list_precip.append(cs)

        chemical_potentials = np.squeeze(wks.eq.MU)

        if len(cs_list_matrix) == 0:
            return None
        
        if any(np.isnan(chemical_potentials)):
            return None
        
        if len(cs_list_precip) > 0:
            miscMatrix = len(cs_list_matrix) > 1
            miscPrec = len(cs_list_precip) > 1
            cs_matrix = cs_list_matrix[0]
            cs_precip = cs_list_precip[0]
            self._compset_cache_df[precPhase] = [cs_matrix, cs_precip]

            if miscMatrix or miscPrec:
                phases, sub_models = self._setupSubModels([self.phases[0], precPhase])
                result, composition_sets = local_equilibrium(self.db, self.elements, phases, cond,
                                                             sub_models, self.phase_records,
                                                             composition_sets=self._compset_cache_df[precPhase])
                self._compset_cache_df[precPhase] = composition_sets
                chemical_potentials = result.chemical_potentials
                
                cs_matrix = [cs for cs in composition_sets if cs.phase_record.phase_name == self.phases[0]][0]
                cs_precip = [cs for cs in composition_sets if cs.phase_record.phase_name == precPhase][0]
        else:
            return None
        
        return chemical_potentials, cs_matrix, cs_precip
    
    def _getCompositionSetsCache(self, x, T, cond, precPhase = None):
        '''
        Gets composition set from x and T by global equilibrium

        Steps
            1. Compute local equilibrium at x and T using previous composition sets
            2. Get composition sets and update cache
                If equilibrium did not converge, then return None
            3. Return values

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)

        Returns
        -------
        chemical_potentials
        cs_matrix - composition set of matrix phase
        cs_precip - composition set of precipitate phase
        x_precip - composition of precipitate phase
        '''
        phases, sub_models = self._setupSubModels([self.phases[0], precPhase])
        result, composition_sets = local_equilibrium(self.db, self.elements, phases, cond,
                                                         sub_models, self.phase_records,
                                                         composition_sets=self._compset_cache_df[precPhase])
        self._compset_cache_df[precPhase] = composition_sets
        chemical_potentials = result.chemical_potentials
        if any(np.isnan(chemical_potentials)):
            return None
        
        ph = [cs.phase_record.phase_name for cs in composition_sets if cs.NP > 0]
        if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
            cs_precip = [cs for cs in composition_sets if cs.phase_record.phase_name == precPhase][0]
            cs_matrix = [cs for cs in composition_sets if cs.phase_record.phase_name == self.phases[0]][0]
        else:
            return None
        
        return chemical_potentials, cs_matrix, cs_precip
    
    def _getPrecCompositionSetSamplingDF(self, x, T, cond, precPhase = None, training = False):
        '''
        Gets samples for precipitate phase for use in sampling driving force method and returns driving force and precipitate composition

        This is also use in tangent driving force method for when equilibrium is not (yet) cached

        Steps
            1. Compute local equilibrium at x and T of only the matrix phase
            2. Sample precipitate phase
                If ordered contribution to matrix phase, then sample ordering contribution
                and remove points on the matrix free energy surface
            3. Compute energy difference between precipitate samples and chemical potential hyperplane
            4. Find sample that maximizes energy difference and return sample composition and driving force

        Parameters
        ----------
        x : float or array
            Composition of minor element in bulk matrix phase
            Use float for binary systems
            Use array for multicomponent systems
        T : float
            Temperature in K
        precPhase : str (optional)
            Precipitate phase to consider (default is first precipitate phase in list)
        returnComp : bool (optional)
            Whether to return composition of precipitate (defaults to False)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        driving force - max free energy difference
        precipitate composition - corresponds to max driving force
        '''
        orderTol = -1e-8

        #Equilibrium at matrix composition for only the parent phase
        phases, sub_models = self._setupSubModels([self.phases[0]])
        self._parentEq, self._matrix_cs = local_equilibrium(self.db, self.elements, phases, cond,
                                                            sub_models, self.phase_records,
                                                            composition_sets = self._matrix_cs)

        if any(np.isnan(self._parentEq.chemical_potentials)):
            return None
        
        #Sample precipitate phase and get driving force differences at all points -------------------------------------------------------------------
        #Sample points of precipitate phase
        phases, sub_models = self._setupSubModels([precPhase])
        if self._pointsPrec[precPhase] is None or self._prevTemperature != T:
            self._pointsPrec[precPhase] = calculate(self.db, self.elements, phases[0], P = 101325, T = T, GE=self.gOffset, pdens = self.sampling_pDens, model=sub_models, output='GM', phase_records=self.phase_records, to_xarray=False)
            if self.orderedPhase[precPhase]:
                self._orderingPoints[precPhase] = calculate(self.db, self.elements, phases[0], P = 101325, T = T, GE=self.gOffset, pdens = self.sampling_pDens, model=sub_models, output='OCM', phase_records=self.phase_records, to_xarray=False)
            self._prevTemperature = T

        #Get value of chemical potential hyperplane at composition of sampled points
        precComp = np.squeeze(self._pointsPrec[precPhase].X)
        mu = np.array([self._parentEq.chemical_potentials])
        mult = precComp * mu

        #Difference between the chemical potential hyperplane and the samples points
        #The max driving force is the same as when the chemical potentials of the two phases are parallel
        diff = np.sum(mult, axis=1) - np.squeeze(self._pointsPrec[precPhase].GM)

        #For phases at fixed composition, there will only be 1 set of site fractions
        #So we make this 2d to properly index
        y = np.atleast_2d(np.squeeze(self._pointsPrec[precPhase].Y))
            
        #Find maximum driving force and corresponding composition -----------------------------------------------------------------------------------
        #For phases with order/disorder transition, a filter is applied such that it will only use points that are below the disordered energy surface
        if self.orderedPhase[precPhase]:
            indices = np.squeeze(self._orderingPoints[precPhase].OCM) < orderTol
            diff = diff[indices]
            y = y[indices]

        dg = np.amax(diff)
        idx = np.argmax(diff)

        prec_cs = CompositionSet(self.phase_records[precPhase])
        state_variables = np.array([cond[v.GE], cond[v.N], cond[v.P], cond[v.T]], dtype=np.float64)
        y = np.array(y[idx][:prec_cs.phase_record.phase_dof], dtype=np.float64)
        prec_cs.update(y, 1, state_variables)

        return dg, prec_cs



from kawin.thermo.Thermodynamics import GeneralThermodynamics
import numpy as np
from pycalphad import Workspace, equilibrium, calculate, variables as v
from pycalphad.core.composition_set import CompositionSet
from kawin.thermo.FreeEnergyHessian import dMudX

class BinaryThermodynamics (GeneralThermodynamics):
    '''
    Class for defining driving force and interfacial composition functions
    for a binary system using pyCalphad and thermodynamic databases

    Parameters
    ----------
    database : str
        File name for database
    elements : list
        Elements to consider
        Note: reference element must be the first index in the list
    phases : list
        Phases involved
        Note: matrix phase must be first index in the list
    drivingForceMethod : str (optional)
        Method used to calculate driving force
        Options are 'tangent' (default), 'approximate', 'sampling', and 'curvature' (not recommended)
    interfacialCompMethod: str (optional)
        Method used to calculate interfacial composition
        Options are 'equilibrium' (default) and 'curvature' (not recommended)
    parameters : list [str] or dict {str : float}
        List of parameters to keep symbolic in the thermodynamic or mobility models
    '''
    def __init__(self, database, elements, phases, drivingForceMethod = 'tangent', interfacialCompMethod = 'equilibrium', parameters = None):
        super().__init__(database, elements, phases, drivingForceMethod, parameters)

        if self.elements[1] < self.elements[0]:
            self.reverse = True
        else:
            self.reverse = False

        #Guess composition for when finding tieline
        self._guessComposition = {self.phases[i]: (0, 1, 0.1) for i in range(1, len(self.phases))}

        self.setInterfacialMethod(interfacialCompMethod)


    def setInterfacialMethod(self, interfacialCompMethod):
        '''
        Changes method for caluclating interfacial composition

        Parameters
        ----------
        interfacialCompMethod - str
            Options are ['equilibrium', 'curvature']
        '''
        if interfacialCompMethod == 'equilibrium':
            self._interfacialComposition = self._interfacialCompositionFromEq
        elif interfacialCompMethod == 'curvature':
            self._interfacialComposition = self._interfacialCompositionFromCurvature
        else:
            raise Exception('Interfacial composition method must be either \'equilibrium\' or \'curvature\'')

    def setGuessComposition(self, conditions):
        '''
        Sets initial composition when calculating equilibrium for interfacial energy

        Parameters
        ----------
        conditions : float, tuple or dict
            Guess composition(s) to solve equilibrium for
            This should encompass the region where a tieline can be found
            between the matrix and precipitate phases
            Options:    float - will set to all precipitate phases
                        tuple - (min, max dx) will set to all precipitate phases
                        dictionary {phase name: scalar or tuple}
        '''
        if isinstance(conditions, dict):
            #Iterating over conditions dictionary in case not all precipitate phases are listed
            for p in conditions:
                self._guessComposition[p] = conditions[p]
        #If not dictionary, then set to all phases
        else:
            for i in range(1, len(self.phases)):
                self._guessComposition[self.phases[i]] = conditions

    def getInterfacialComposition(self, T, gExtra = 0, precPhase = None):
        '''
        Gets interfacial composition accounting for Gibbs-Thomson effect

        Parameters
        ----------
        T : float or array
            Temperature in K
        gExtra : float or array (optional)
            Extra contributions to the precipitate Gibbs free energy
            Gibbs Thomson contribution defined as Vm * (2*gamma/R + g_Elastic)
            Defaults to 0
        precPhase : str
            Precipitate phase to consider (default is first precipitate in list)

        Note: for multiple conditions, only gExtra has to be an array
            This will calculate compositions for multiple gExtra at the input Temperature

            If T is also an array, then T and gExtra must be the same length
            where each index will pertain to a single condition

        Returns
        -------
        (parent composition, precipitate composition)
        Both will be either float or array based off shape of gExtra
        Will return (None, None) if precipitate is unstable
        '''
        if hasattr(gExtra, '__len__'):
            if not hasattr(T, '__len__'):
                caArray, cbArray = self._interfacialComposition(T, gExtra, precPhase)
            else:
                #If T is also an array, then iterate through T and gExtra
                #Otherwise, pycalphad will create a cartesian product of the two
                caArray = []
                cbArray = []
                for i in range(len(gExtra)):
                    ca, cb = self._interfacialComposition(T[i], gExtra[i], precPhase)
                    caArray.append(ca)
                    cbArray.append(cb)
                caArray = np.array(caArray)
                cbArray = np.array(cbArray)

            return caArray, cbArray
        else:
            return self._interfacialComposition(T, gExtra, precPhase)


    def _interfacialCompositionFromEq(self, T, gExtra = 0, precPhase = None):
        '''
        Gets interfacial composition by calculating equilibrum with Gibbs-Thomson effect

        Parameters
        ----------
        T : float
            Temperature in K
        gExtra : float (optional)
            Extra contributions to the precipitate Gibbs free energy
            Gibbs Thomson contribution defined as Vm * (2*gamma/R + g_Elastic)
            Defaults to 0
        precPhase : str
            Precipitate phase to consider (default is first precipitate in list)

        Returns
        -------
        (parent composition, precipitate composition)
        Both will be either float or array based off shape of gExtra
        Will return (None, None) if precipitate is unstable
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        if hasattr(gExtra, '__len__'):
            gExtra = np.array(gExtra)
        else:
            gExtra = np.array([gExtra])
        gExtra += self.gOffset

        #Compute equilibrium at guess composition
        cond = {v.X(self.elements[1]): self._guessComposition[precPhase], v.T: T, v.P: 101325, v.GE: gExtra}
        phases, sub_models = self._setupSubModels(precPhase)
        wks = Workspace(self.db, self.elements, phases, cond, models=sub_models,
                        phase_record_factory = self.phase_records, calc_opts={'pdens': self.pDens})
        
        coord_keys = list(wks.eq.coords.keys())
        ge_var_idx = coord_keys.index('GE')

        xMatrixArray = -1*np.ones(len(gExtra), dtype=np.float64)
        xPrecipArray = -1*np.ones(len(gExtra), dtype=np.float64)
        gIndex = 0
        # cs_idx is a 5 len tuple of (GE, N, P, T, X)
        # where it iterates by X first, then by GE
        for cs_idx, cs_list in wks.enumerate_composition_sets():
            # If the GE index is greater than the current record index, then
            # update the current index. This occurs if we did not find any
            # two-phase regions at a given value of GE and now we want to move
            # on to the next value of GE
            if cs_idx[ge_var_idx] > gIndex:
                gIndex = cs_idx[ge_var_idx]

            # Only check for two-phase region if the cs list corresponds
            # to the current index of GE that we're looking at
            # If the current index of GE is greater than what's in
            # cs_idx, then we want to iterate the composition sets until we
            # catch up to gIndex
            if cs_idx[ge_var_idx] == gIndex:
                ph = [cs.phase_record.phase_name for cs in cs_list]
                if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
                    cs_matrix = [cs for cs in cs_list if cs.phase_record.phase_name == self.phases[0]][0]
                    cs_precip = [cs for cs in cs_list if cs.phase_record.phase_name == precPhase][0]

                    c_idx = 0 if self.reverse else 1
                    xMatrixArray[gIndex] = cs_matrix.X[c_idx]
                    xPrecipArray[gIndex] = cs_precip.X[c_idx]
                    gIndex += 1

        if len(gExtra) == 1:
            return xMatrixArray[0], xPrecipArray[0]
        else:
            return xMatrixArray, xPrecipArray

    def _interfacialCompositionFromCurvature(self, T, gExtra = 0, precPhase = None):
        '''
        Gets interfacial composition using free energy curvature
        G''(x - xM)(xP-xM) = 2*y*V/R

        Parameters
        ----------
        T : float
            Temperature in K
        gExtra : float (optional)
            Extra contributions to the precipitate Gibbs free energy
            Gibbs Thomson contribution defined as Vm * (2*gamma/R + g_Elastic)
            Defaults to 0
        precPhase : str
            Precipitate phase to consider (default is first precipitate in list)

        Returns
        -------
        (parent composition, precipitate composition)
        Both will be either float or array based off shape of gExtra
        Will return (None, None) if precipitate is unstable
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        if hasattr(gExtra, '__len__'):
            gExtra = np.array(gExtra)
        else:
            gExtra = np.array([gExtra])

        #Compute equilibrium at guess composition
        cond = {v.X(self.elements[1]): self._guessComposition[precPhase], v.T: T, v.P: 101325, v.GE: self.gOffset, v.N: 1}
        phases, sub_models = self._setupSubModels(precPhase)
        wks = Workspace(self.db, self.elements, phases, cond, models=sub_models,
                        phase_record_factory = self.phase_records, calc_opts={'pdens': self.pDens})
        
        chemical_potentials = np.squeeze(wks.eq.MU)

        idx = 0
        for _, cs_list in wks.enumerate_composition_sets():
            ph = [cs.phase_record.phase_name for cs in cs_list]
            if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
                cs_matrix = [cs for cs in cs_list if cs.phase_record.phase_name == self.phases[0]][0]
                cs_precip = [cs for cs in cs_list if cs.phase_record.phase_name == precPhase][0]
                chem_pot = chemical_potentials[idx]

                dMudxMatrix = dMudX(chem_pot, cs_matrix, self.elements[0])[0,0]
                dMudxPrecip = dMudX(chem_pot, cs_precip, self.elements[0])[0,0]

                c_idx = 0 if self.reverse else 1
                xMatrixEq = cs_matrix.X[c_idx]
                xPrecipEq = cs_precip.X[c_idx]

                if dMudxMatrix != 0:
                    xMatrix = gExtra / dMudxMatrix / (xPrecipEq - xMatrixEq) + xMatrixEq
                else:
                    xMatrix = xMatrixEq*np.ones(len(gExtra))

                if dMudxPrecip != 0:
                    xPrecip = dMudxMatrix * (xMatrix - xMatrixEq) / dMudxPrecip + xPrecipEq
                else:
                    xPrecip = xPrecipEq*np.ones(len(gExtra))

                xMatrix[xMatrix < 0] = 0
                xMatrix[xMatrix > 1] = 1
                xPrecip[xPrecip < 0] = 0
                xPrecip[xPrecip > 1] = 1

                if len(gExtra) == 1:
                    return xMatrix[0], xPrecip[0]
                else:
                    return xMatrix, xPrecip
            idx += 1

        if len(gExtra) == 1:
            return -1, -1
        else:
            return -1*np.ones(len(gExtra)), -1*np.ones(len(gExtra))

    def plotPhases(self, ax, T, gExtra = 0, plotGibbsOffset = False, *args, **kwargs):
        '''
        Plots sampled points from the parent and precipitate phase

        Parameters
        ----------
        ax : Axis
        T : float
            Temperature in K
        gExtra : float (optional)
            Extra contributions to the Gibbs free energy of precipitate
            Defaults to 0
        plotGibbsOffset : bool (optional)
            If True and gExtra is not 0, the sampled points of the
                precipitate phase will be plotted twice with gExtra and
                with no extra Gibbs free energy contributions
            Defualts to False
        '''
        phases, sub_models = self._setupSubModels([self.phases[0]])
        points = calculate(self.db, self.elements, phases[0], P=101325, T=T, GE=0, model=sub_models, phase_records=self.phase_records, output='GM')
        ax.scatter(points.X.sel(component=self.elements[1]), points.GM / 1000, label=self.phases[0], *args, **kwargs)

        #Add gExtra to precipitate phase
        for i in range(1, len(self.phases)):
            phases, sub_models = self._setupSubModels([self.phases[i]])
            points = calculate(self.db, self.elements, phases[0], P=101325, T=T, GE=0, model=sub_models, phase_records=self.phase_records, output='GM')
            ax.scatter(points.X.sel(component=self.elements[1]), (points.GM + gExtra) / 1000, label=self.phases[i], *args, **kwargs)

            #Plot non-offset precipitate phase
            if plotGibbsOffset and gExtra != 0:
                ax.scatter(points.X.sel(component=self.elements[1]), points.GM / 1000, color='silver', alpha=0.3, *args, **kwargs)

        ax.legend()
        ax.set_xlim([0, 1])
        ax.set_xlabel('Composition ' + self.elements[1])
        ax.set_ylabel('Gibbs Free Energy (kJ/mol)')